#include <easy.h>

#undef E_DEBUG

e_obj_descriptor e_obj_descriptor_aset = {
  "aset",             /* Name      */
  0,                  /* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_aset_free,        /* free      */
  e_aset_to_string,   /* to_string */
  e_aset_clone,       /* clone     */
  e_aset_equals,      /* equals    */
};

e_obj_descriptor e_obj_descriptor_aset_item = {
  "aset_item",        /* Name      */
  0,                  /* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_aset_item_free,   /* free      */
  e_aset_item_to_string, /* to_string */
  e_aset_item_clone,  /* clone     */
  e_aset_item_equals, /* equals    */
};

/************************************************************/
/* Aset :                                                   */
/* Binary tree or aset                                      */
/************************************************************/
/***************************************************************************/
/**
 *  e_aset *e_aset_new(e_aset *asp)
 *                    [0;+1]
 **/
/***************************************************************************/
/* Create or initialise a new e_aset                                       */
/* asp : e_aset to initialise (static e_aset)                              */
/*      if NULL allocate it                                                */
/*      asp is assumed to be of the correct type                           */
/* Return e_aset                                                           */
/*        0 : error                                                        */
/* Link : return : +1                                                      */
/***************************************************************************/
e_aset *e_aset_new(e_aset *asp)
{ 
#ifdef E_DEBUG
  e_debug("e_aset_new(asp=%x)\n",asp);
#endif
                 /* Static object or allocate it ? */
  if (!asp)
  { if (!(asp = e_obj_descriptor_aset._alloc(sizeof(e_aset)))) 
      return((e_aset *)e_error_no_memory("e_aset_new"));
    asp->h.refs = 1;
  }
  else asp->h.refs = -1;
  asp->h.mode = 0;

  asp->h.obj_desc_p = &e_obj_descriptor_aset;

  E_ASET_FIRST(asp) = 0;  
  E_ASET_LAST(asp) = 0;  
  E_ASET_CUR(asp) = 0;  
  E_ASET_CARDINAL(asp) = 0;

  return(asp);
}

/***************************************************************************/
/**
 *  e_aset_item *e_aset_item_new(e_aset_item *aip,e_name *np,void *val)
 *                              [0,-1,-1;+1]
 **/
/***************************************************************************/
/* Create or initialise a new e_aset_item                                  */
/* aip : e_aset_item to initialise (static e_aset)                         */
/*      if NULL allocate it                                                */
/*      aip is assumed to be of the correct type                           */
/* np  : name for the item (can be null)                                   */
/* val : value for the item (can be null)                                  */
/* Return e_aset_item                                                      */
/*        0 : error                                                        */
/* Link : np : -1                                                          */
/*        val : -1                                                         */
/*        Return : +1                                                      */
/***************************************************************************/
e_aset_item *e_aset_item_new(e_aset_item *aip,e_name *np,void *val)
{ 
#ifdef E_DEBUG
  e_debug("e_aset_item_new(aip=%x,np=%x,val=%x)\n",aip,np,val);
#endif
 
                 /* Static object or allocate it ? */
  if (!aip)
  { if (!(aip = e_obj_descriptor_aset._alloc(sizeof(e_aset_item)))) 
      return((e_aset_item *)e_error_no_memory("e_aset_item_new"));
    aip->h.refs = 1;
  }
  else aip->h.refs = -1;
  aip->h.mode = 0;

  aip->h.obj_desc_p = &e_obj_descriptor_aset_item;

  E_ASET_ITEM_PREV(aip) = 0;
  E_ASET_ITEM_NEXT(aip) = 0;
  E_ASET_ITEM_NAME(aip) = np;  
  E_ASET_ITEM_VAL(aip) = val;  

  return(aip);
}

/***************************************************************************/
/**
 *  void e_aset_free(void *asp)
 *                   [-1;.]
 **/
/***************************************************************************/
/* Free the e_aset                                                         */
/* If the refs counter is -1 : static aset : leave as is                   */
/* If the refs counter is 1 : free the allocated memory for the e_aset     */
/* Otherwise decrement the counter by 1.                                   */
/* REF : stp : -1                                                          */
/***************************************************************************/
void e_aset_free(void *asp)
{ e_aset_item *aip,*next;

#ifdef E_DEBUG
  e_debug("e_aset_free(asp=%x)\n",asp);
#endif

  if(asp)
  { if(E_OBJ_REF(asp) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free aset (%x) ",asp);
      goto error;
    }
                                /* static aset */
    if (E_OBJ_REF(asp) < 0) return;

                                /* decrement the refs counter   */
    if (E_OBJ(asp)->refs-- > 1) return;

                                /* Last reference,     
                                   Free the two element, 
                                    then free the aset */
                 /* Free each item                        */
                 /* items can't be in more than one list. */
                 /* Their Ref counter should always be 1  */
    aip = E_ASET_FIRST(asp);
    while(aip)
    { next = E_ASET_ITEM_NEXT(aip);
      e_aset_item_free(aip);
      aip = next;
    }

    e_obj_descriptor_aset._free(asp);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/***************************************************************************/
/**
 *  void e_aset_item_free(void *aip)
 *                   [-1;.]
 **/
/***************************************************************************/
/* Free the e_aset_item                                                    */
/* e_aset_item  can only be inserted into 1 list, so their Ref counter     */
/* Must be one. This should only be called by functions manipulating       */
/* asets. They must deal with the pointers.                                */
/* The Name and value of the aset_item are freed (and the the e_aset_item) */
/* REF : aip : -1                                                          */
/***************************************************************************/
void e_aset_item_free(void *aip)
{ int free = 1;

#ifdef E_DEBUG
  e_debug("e_aset_item_free(aip=%x)\n",aip);
#endif

  if(aip)
  { if(E_OBJ_REF(aip) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free aset_item (%x) ",aip);
      goto error;
    }
                                /* static aset */
    if (E_OBJ_REF(aip) < 0) 
    { free = 0; goto end;
    }
                                /* decrement the refs counter   */
    if (E_OBJ(aip)->refs != 1) 
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "aset_item has ref counter larger than 1 : (%x) ",aip);
      goto error;
    }

end:
                   /* Free name and value */
    if(E_ASET_ITEM_NAME(aip)) E_FREE(E_ASET_ITEM_NAME(aip));
    if(E_ASET_ITEM_VAL(aip)) E_FREE(E_ASET_ITEM_VAL(aip));

    if(free) e_obj_descriptor_aset_item._free(aip);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/*************************************************************/
/**
 *  void *e_aset_to_string(void *asp,int tab,e_flag mode)
 *                         [0,.,.;+1]
 **/
/*************************************************************/
/* Convert the aset to a string                              */
/* aip : an e_aset, assumed to be of the correct type        */
/* { Name1 = Value;                                          */
/*   Name2 = Value1;                                         */
/*   ...                                                     */
/* }                                                         */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : asp : 0                                            */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_aset_to_string(void *asp,int tab,e_flag mode)
{ e_string *stp = 0,*stp2 = 0;
  e_aset_item *aip = 0;
  char *s;

#ifdef E_DEBUG
  e_debug("e_aset_to_string(asp=%x,tab=%d,mode=%x)\n",asp,tab,mode);
#endif
                              /* write { before the first element */

  if(!E_ASET_CARDINAL(asp))  /* Empty set : {} */
  { E_NULL_ERROR(stp = e_string_new(0,"{ }",tab+5));

    if(!(mode & E_TO_STR_NO_TAB))
    { stp = e_string_insert_chars(stp,0,tab,' '); /* add tab to the string */
    }

    return(stp);
  }

  aip = E_ASET_FIRST(asp);
  E_NULL_ERROR(stp = e_string_new(0,"{ ",E_LINE_SIZE));
                             /* First item : same line as { */

  E_NULL_ERROR(stp = e_string_append_string(stp,(e_string *)
             e_aset_item_to_string(aip,tab+2,mode|E_TO_STR_NO_TAB_ON_FIRST))
              ); 

  if(mode & E_TO_STR_COMPACT) s = ";";
  else                        s = ";\n";
  E_NULL_ERROR(stp = e_string_append_buffer(stp,s));

  if(!(mode & E_TO_STR_NO_TAB))    
  { stp = e_string_insert_chars(stp,0,tab,' ');   /* add tab to the string */
  }

  while((aip = aip->next))
  { E_NULL_ERROR(stp2 = E_TO_STRING(aip,tab+2,
                                    mode&(~E_TO_STR_NO_TAB_ON_FIRST)));
    if(mode & E_TO_STR_COMPACT) s = ";";
    else                        s = ";\n";
    E_NULL_ERROR(stp2 = e_string_append_buffer(stp2,s));
    if(!(stp = e_string_append_string(stp,stp2))) goto error2;
  }

  E_NULL_ERROR(stp2 = e_string_new(0,"}",3));

  if(!(mode & E_TO_STR_COMPACT))
  { stp2 = e_string_insert_chars(stp2,0,tab,' ');   /* add tab to the string */
  }
  if(!(stp = e_string_append_string(stp,stp2))) goto error2;

  return (stp);

error:
  if (stp2) E_FREE(stp2);
error2:
  if (stp) E_FREE(stp);
  return(0);
}

/*************************************************************/
/**
 *  void *e_aset_item_to_string(void *aip,int tab,e_flag mode)
 *                              [0,.,.;+1]
 **/
/*************************************************************/
/* Convert the aset_item to a string                         */
/* Name = Value or Value                                     */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : aip : 0                                            */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_aset_item_to_string(void *aip,int tab,e_flag mode)
{ e_string *stp = 0,*stp2 = 0;
  
#ifdef E_DEBUG
  e_debug("e_aset_item_to_string(aip=%x,tab=%d,mode=%x)\n",aip,tab,mode);
#endif

                                    /* Name = */
  if(E_ASET_ITEM_NAME(aip))
  { E_NULL_ERROR(stp = E_TO_STRING(E_ASET_ITEM_NAME(aip),0,mode));
    E_NULL_ERROR(stp = e_string_append_buffer(stp," = "));
  }
                                    /* ... Value */
  if(E_ASET_ITEM_VAL(aip))
  { E_NULL_ERROR(stp2 = E_TO_STRING(E_ASET_ITEM_VAL(aip),tab,
                                    mode|E_TO_STR_NO_TAB_ON_FIRST));
  }

  if(stp) { E_NULL_ERROR(stp = e_string_append_string(stp,stp2)); }
  else stp = stp2;

  if(!(mode & E_TO_STR_NO_TAB))
  { stp = e_string_insert_chars(stp,0,tab,' '); /* add tab to the string */
  }
  return(stp);

error:
  if (stp) E_FREE(stp);
  return(0);
}

/*************************************************************/
/**
 *   void *e_aset_clone(void *asp,e_flag mode)
 *                     [0,.;+1]
 **/
/*************************************************************************/
/* Clone an aset                                                         */
/* mode : cloning mode                                                   */
/* If the aset can't be cloned, return asp and increase its refs counter */
/* asp : an aset, assumed to be of the correct type                      */
/* mode : E_CLONE_RECURSIVE : to clone the object recursively            */
/* Return : cloned e_aset                                                */
/*          0 : Error                                                    */
/* LINK : asp : 0                                                        */
/*        RETURN : +1                                                    */
/*************************************************************************/
void *e_aset_clone(void *asp,e_flag mode)
{ e_aset *new_asp=0;
  e_aset_item *aip,*new_aip=0,*aip_last=0;

#ifdef E_DEBUG
  e_debug("e_aset_clone(asp=%x,mode=%x)\n",asp,mode);
#endif

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(asp)) return(E_USE(asp));

  E_NULL_ERROR(new_asp = e_aset_new(0));

              /* The aset is always cloned recursively */
  aip = E_ASET_FIRST(asp);
  while(aip)
  { E_NULL_ERROR(new_aip = e_aset_item_clone(aip,mode));
    if(!aip_last) E_ASET_FIRST(new_asp) = new_aip;
    else E_ASET_ITEM_NEXT(aip_last) = new_aip;
    E_ASET_ITEM_PREV(new_aip) = aip_last;
    
              /* Clone cur item pointer */
    if(aip == E_ASET_CUR(asp)) E_ASET_CUR(new_asp) = new_aip;

    aip_last = new_aip;
    aip = E_ASET_ITEM_NEXT(aip);
  }
  E_ASET_LAST(new_asp) = new_aip;
  E_ASET_CARDINAL(new_asp) = E_ASET_CARDINAL(asp);

  return(new_asp);

error:
  if(new_asp)
  { if((aip = E_ASET_FIRST(new_asp)))
    { while(aip) 
      { new_aip = E_ASET_ITEM_NEXT(aip);
        e_aset_item_free(aip);
        aip = new_aip;
      }
    } 
    e_obj_descriptor_aset_item._free(new_asp);
  }  
  return(0);
}

/*************************************************************/
/**
 *   void *e_aset_item_clone(void *aip,e_flag mode)
 *                          [0,.;+1]
 **/
/*************************************************************************/
/* Clone an aset_item                                                    */
/* mode : cloning mode                                                   */
/* The next and prev pointers are not initialised, only the names and    */
/* the values are cloned.                                                */
/* aip : an aset_item, assumed to be of the correct type                 */
/* mode : E_CLONE_RECURSIVE : to clone the object recursively            */
/* Return : cloned e_aset                                                */
/*          0 : Error                                                    */
/* LINK : asp : 0                                                        */
/*        RETURN : +1                                                    */
/*************************************************************************/
void *e_aset_item_clone(void *aip,e_flag mode)
{ e_aset_item *new_aip;
  e_name *np;
  void *val;

#ifdef E_DEBUG
  e_debug("e_aset_item_clone(aip=%x,mode=%x)\n",aip,mode);
#endif

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(aip)) return(E_USE(aip));

  if((np = E_ASET_ITEM_NAME(aip))) E_OBJ_REF(np)++;  
  val = E_ASET_ITEM_VAL(aip);  

                 /* Clone val or increase counter ? */
  if(mode & E_CLONE_RECURSIVE) val = E_CLONE(val,mode);
  else                         val = E_USE(val);

  E_NULL_ERROR(new_aip = e_aset_item_new(0,np,val));

  return(new_aip);

error:
  if(val && (mode & E_CLONE_RECURSIVE)) E_FREE(val);
  if(np) E_FREE(np);
  return(0);
}

/*************************************************************/
/**
 *   int e_aset_equals(void *p1,void *p2)
 *                    [0,0;.]
 **/
/***************************************************************/
/* Test if the 2 asets are equal:                              */
/* The two asets are equal if they are identical (p1 == p2),   */
/* If they have the same cardinal and identical elements.      */
/* The current item is not checked!                            */
/* Return : 1 : p1 and p2 are equal                            */
/*          0 : p1 and p2 are different                        */
/* Link : p1 0                                                 */
/*        p2 0                                                 */
/***************************************************************/
int e_aset_equals(void *p1,void *p2)
{ e_aset_item *aip1,*aip2;

#ifdef E_DEBUG
  e_debug("e_aset_equals(p1=%x,p2=%x)\n",p1,p2);
#endif

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

            /* different objet  types => differ */
  if(E_OBJ_DESCR(p1) != E_OBJ_DESCR(p2)) return(0);

            /* compare the cradinals */
  if(E_ASET_CARDINAL(p1) != E_ASET_CARDINAL(p2)) return(0);

  aip1 = E_ASET_FIRST(p1);
  aip2 = E_ASET_FIRST(p2);
  while(aip1 && aip2)
  { if(!e_aset_item_equals(aip1,aip2)) return(0);
    aip1 = E_ASET_ITEM_NEXT(aip1);
    aip2 = E_ASET_ITEM_NEXT(aip2);
  }

  return(1);
}

/*************************************************************/
/**
 *   int e_aset_item_equals(void *p1,void *p2)
 *                         [0,0;.]
 **/
/***************************************************************/
/* Test if the 2 aset_items are equal:                         */
/* The two asets_items are equal if they are identical (p1==p2)*/
/* or if they have the same value and the same name            */
/* Return : 1 : p1 and p2 are equal                            */
/*          0 : p1 and p2 are different                        */
/* Link : p1 0                                                 */
/*        p2 0                                                 */
/***************************************************************/
int e_aset_item_equals(void *p1,void *p2)
{ void *val1,*val2;

#ifdef E_DEBUG
  e_debug("\e_aset_item_equals(p1=%x,p2=%x)n",p1,p2);
#endif

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

  if(!e_name_equals(E_ASET_ITEM_NAME(p1),E_ASET_ITEM_NAME(p2))) return(0);

  val1 = E_ASET_ITEM_VAL(p1);
  val2 = E_ASET_ITEM_VAL(p2);

  if(val1 == val2) return(1);  
  if(!val1) return(0);

  return(E_EQUALS(val1,val2));
}

/*************************************************************/
/**
 *   int e_aset_next(void *asp,int mode,void **obj)
 *                  [0,*+1;.]
 **/
/*********************************************************************/
/* Get the next item value from the aset and write it in obj.        */
/* asp : aset to scan : supposed to be of type e_aset.               */
/* asp.cur is - FIRST when refering to the first element.            */
/*            - 0 when the end of set has been reached.              */
/* mode : E_NEXT_GET_NEXT   : (0) get next item value                */
/*        E_NEXT_GET_CUR : get the current item value,               */
/*                         the counter is left unchanged.            */
/*        E_NEXT_GET_NEXT_KEY : get the next item key name           */
/*        E_NEXT_GET_CUR_KEY  : get the current item key name        */
/*                         the counter is left unchanged.            */
/*        E_NEXT_RESET : reset index counter to the first element    */
/*        E_NEXT_NEXT : move to next item. **obj set to 0 if non NULL*/
/* obj : pointer to the object pointer variable: in which the object */
/*       must be written into with a link                            */
/* Return : 1 : OK                                                   */
/*          -1 : no more elements                                    */
/*          0 : error or no more element                             */
/* Link : asp : 0                                                    */
/*        *obj : +1                                                  */
/*********************************************************************/
int e_aset_next(void *asp,int mode,void **obj)
{ 

#ifdef E_DEBUG
  e_debug("e_aset_next(asp=%x,mode=%x,obj=%x)\n",asp,mode,obj);
#endif

  if(!asp) { e_error_null_pointer("e_aset_next(asp,-,-)"); goto error;}

  if(mode == E_NEXT_RESET) /* Reset the aset counter */
  { E_ASET_CUR(asp) = E_ASET_FIRST(asp);
    return(1);
  }

  if(mode == E_NEXT_NEXT) /* Reset the aset counter */
  { if(E_ASET_CUR(asp)) E_ASET_CUR(asp) = E_ASET_ITEM_NEXT(E_ASET_CUR(asp));
    if(obj) *obj = 0;
    return(1);
  }

  if(!obj) { e_error_null_pointer("e_aset_next(-,-,obj)"); goto error;}

  if(E_ASET_CUR(asp) == 0) /* end of aset */
  { return(-1);
  }

  switch(mode)
  { case E_NEXT_GET_NEXT :
    case E_NEXT_GET_CUR :
           if((*obj = E_ASET_ITEM_VAL(E_ASET_CUR(asp)))) E_REF(*obj);
           break;
    case E_NEXT_GET_NEXT_KEY :
    case E_NEXT_GET_CUR_KEY :
           if((*obj = E_ASET_ITEM_NAME(E_ASET_CUR(asp)))) E_REF(*obj);
           break;
    default : 
           break;
  } 

  if((mode == E_NEXT_GET_NEXT)||(mode == E_NEXT_GET_NEXT_KEY))
  { E_ASET_CUR(asp) = E_ASET_ITEM_NEXT(E_ASET_CUR(asp));
  }

  return(1); 

error:
  
  return(0);
}

/************************ ADD TO ASET *************************************/

/******************************************************************/
/**
 *   e_aset *e_aset_append_item(e_aset *asp,e_name *np, void *val)
 *                             [-1,-1,-1;+1]
 **/
/*************************************************/
/* Add one item at the end of the aset asp.      */
/* If the aset is used more than 1, it is cloned */
/* and the clone is modified                     */
/* asp : aset to add the item to.                */
/* np : name for the item (can be NULL)          */
/* val : value for the new item (Can't be NULL)  */
/* Return : the modified aset.                   */
/*          0 : error                            */
/* Link : asp -1                                 */
/*        np  -1                                 */
/*        val -1                                 */
/*        Return : 1                             */
/*************************************************/
e_aset *e_aset_append_item(e_aset *asp,e_name *np, void *val)
{ e_aset *p;
  e_aset_item *aip = 0,*last;

#ifdef E_DEBUG
  e_debug("e_aset_append_item(asp=%x,np=%x,val=%x)\n",asp,np,val);
#endif

  if(!asp) { e_error_null_pointer("e_aset_append_item(asp,-,-)"); goto error;}
  if(!val) { e_error_null_pointer("e_aset_append_item(-,-,val)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

             /* Allocate the new item */ 
  aip = e_aset_item_new(0,np,val);
  val = 0; np = 0; 
  E_NULL_ERROR(aip);

             /* Merge the item at the end of the list */
  if(!(last = E_ASET_LAST(asp)))
  {          /* The aset is empty => aip is the only element */
    E_ASET_LAST(asp) = aip;
    E_ASET_FIRST(asp) = aip;
    E_ASET_CARDINAL(asp) = 1;
  }
  else       /* add item to end of list */
  { E_ASET_ITEM_NEXT(last) = aip;
    E_ASET_ITEM_PREV(aip) = last;
    E_ASET_LAST(asp) = aip;
    E_ASET_CARDINAL(asp)++;
  }  
  return(asp);

error:
  if (asp) E_FREE(asp);  
  if (aip) E_FREE(aip);  
  if (np) E_FREE(np);  
  if (val) E_FREE(val);  
  return(0);
}

/******************************************************************/
/**
 *   e_aset *e_aset_prepend_item(e_aset *asp,e_name *np, void *val)
 *                              [-1,-1,-1;+1]
 **/
/*************************************************/
/* Add one item at the beginning of to the aset  */
/* If the aset is used more than 1, it is cloned */
/* and the clone is modified                     */
/* asp : aset to add the item to.                */
/* np : name for the item (can be NULL)          */
/* val : value for the new item (Can't be NULL)  */
/* Return : the modified aset.                   */
/*          0 : error                            */
/* Link : asp -1                                 */
/*        np  -1                                 */
/*        val -1                                 */
/*        Return : 1                             */
/*************************************************/
e_aset *e_aset_prepend_item(e_aset *asp,e_name *np, void *val)
{ e_aset *p;
  e_aset_item *aip = 0,*first;

#ifdef E_DEBUG
  e_debug("e_aset_prepend_item(asp=%x,np=%x,val=%x)\n",asp,np,val);
#endif

  if(!asp) { e_error_null_pointer("e_aset_append_item(asp,-,-)"); goto error;}
  if(!val) { e_error_null_pointer("e_aset_append_item(-,-,val)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

             /* Allocate the new item */ 
  aip = e_aset_item_new(0,np,val);
  val = 0; np = 0; 
  E_NULL_ERROR(aip);

             /* Add the item at the beginning of the list */
  if(!(first = E_ASET_LAST(asp)))
  {          /* The aset is empty => aip is the only element */
    E_ASET_LAST(asp) = aip;
    E_ASET_FIRST(asp) = aip;
    E_ASET_CARDINAL(asp) = 1;
  }
  else       /* add item to end of list */
  { E_ASET_ITEM_PREV(first) = aip;
    E_ASET_ITEM_NEXT(aip) = first;
    E_ASET_FIRST(asp) = aip;
    E_ASET_CARDINAL(asp)++;
  }  
  return(asp);

error:
  if (asp) E_FREE(asp);  
  if (aip) E_FREE(aip);  
  if (np) E_FREE(np);  
  if (val) E_FREE(val);  
  return(0);
}

/*****************************************************************************/
/**
 *   e_aset *e_aset_add_item(e_aset *asp,e_name *np, void *val,int mode)
 *                          [-1,-1,-1,.;+1]
 **/
/*****************************************************************************/
/* Add one item to the aset asp. The item is normally added at the end       */
/* unless the select mode specifies otherwise.                               */
/* If the aset is used more than 1, it is cloned and the clone is modified   */
/* asp : aset to add the item to.                                            */
/* np : name for the item key (can be NULL)                                  */
/* val : value for the new item (Can't be NULL)                              */
/* mode : E_ASET_ADD_MODE_UNIQUE_VAL : add the object if it is not in the    */
/*      set yet (usualy the key will be null for this but does not have to)  */
/*      If np is null the name of the object found is not changed.           */
/*      E_ASET_ADD_MODE_UNIQUE_KEY : add the object if the key is not in     */
/*      the set yet. If it is val replaces the old value unless the          */
/*      NO_REPLACE mode is selected too. when both UNIQUE_VAL and UNIQUE_KEY */
/*      are selected UNIQUE_VAL is ignored unless np is NULL.                */
/*      E_ASET_ADD_MODE_NO_REPLACE : when used with mode UNIQUE_KEY do not   */
/*      change the value for the key, and return the set as it was (no error)*/
/* Items that are write protected are not modified (no error generated)      */
/* Return : the modified aset.                                               */
/*          0 : error                                                        */
/* Link : asp -1                                                             */
/*        np  -1                                                             */
/*        val -1                                                             */
/*        Return : 1                                                         */
/*****************************************************************************/
e_aset *e_aset_add_item(e_aset *asp,e_name *np, void *val,int mode)
{ e_aset *p;
  e_aset_item *aip = 0;

#ifdef E_DEBUG
  e_debug("e_aset_add_item(asp=%x,np=%x,val=%x,mode=%x)\n",asp,np,val,mode);
#endif

  if(!asp) 
  { e_error_null_pointer("e_aset_append_item(asp,-,-,-)"); goto error;}
  if(!val) 
  { e_error_null_pointer("e_aset_append_item(-,-,val,-)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

                  /* Add item ensuring the key remains unique */
  if((mode & E_ASET_ADD_MODE_UNIQUE_KEY) && np)
  { 
    if(!(aip = e_aset_get_item_by_key(asp,E_USE(np),1)))
    {                      /* Key not found : add  key = val  */
      return(e_aset_append_item(asp,np,val));
    }

    if(mode & E_ASET_ADD_MODE_NO_REPLACE)
    {                   /* Leave key value as it is */
      E_FREE(np); E_FREE(val);
      return(asp);
    }

    if(!E_OBJ_READ_ONLY(aip))
    { if(E_ASET_ITEM_VAL(aip)) E_FREE(E_ASET_ITEM_VAL(aip));
      E_ASET_ITEM_VAL(aip) = val;
    }
    return(asp);
  }

                  /* Add item ensuring the value remains unique */
  if(mode & E_ASET_ADD_MODE_UNIQUE_VAL)
  { 
    if(!(aip = e_aset_get_item_by_value(asp,E_USE(val))))
    {                      /* object not found : add  key = val  */
      return(e_aset_append_item(asp,np,val));
    }

    if(!E_OBJ_READ_ONLY(aip))
    { if(E_ASET_ITEM_VAL(aip)) E_FREE(E_ASET_ITEM_VAL(aip));
      E_ASET_ITEM_VAL(aip) = val;
      if(np)
      { if(E_ASET_ITEM_NAME(aip)) E_FREE(E_ASET_ITEM_NAME(aip));
        E_ASET_ITEM_NAME(aip) = np;
      }
    }
    return(asp);
  }

                    /* By default simply append the item */
  return(e_aset_append_item(asp,np,val));

error:
  if (asp) E_FREE(asp);  
  if (np) E_FREE(np);  
  if (val) E_FREE(val);  
  return(0);
}

/*****************************************************************************/
/**
 * e_aset *e_aset_add_aset(e_aset *asp,e_name *aspv,int mode)
 *                        [-1,-1,.;+1]
 **/
/*****************************************************************************/
/* Add a set to the aset asp. The items are normally added at the end        */
/* unless the select mode specifies otherwise.                               */
/* If the aset is used more than 1, it is cloned and the clone is modified   */
/* asp : aset to add the item to.                                            */
/* aspv : aset to add. Its items are added one by one                        */
/* mode : Same as for e_aset_add_item but it aplies for each element of aspv.*/
/* Return : the modified aset.                                               */
/*          0 : error                                                        */
/* Link : asp  -1                                                            */
/*        aspv -1                                                            */
/*        Return : 1                                                         */
/*****************************************************************************/
e_aset *e_aset_add_aset(e_aset *asp,e_aset *aspv,int mode)
{ e_aset_item *aip;    
  void *v;
  e_name *np;

#ifdef E_DEBUG
  e_debug("e_aset_add_aset(asp=%x,aspv=%x,mode=%x)\n",asp,aspv,mode);
#endif

  aip = E_ASET_FIRST(aspv);
  while(aip)
  { v = E_ASET_ITEM_VAL(aip);
    if((np = E_ASET_ITEM_NAME(aip))) E_USE(np);

    asp = e_aset_add_item(asp,np,E_USE(v),mode);
    aip = E_ASET_ITEM_NEXT(aip);
  }
  E_FREE(aspv);
  return(asp);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_append(e_aset *asp,e_aset *asp2)
 *                        [-1,-1;+1]
 **/
/*************************************************/
/* Append one aset at the end of asp.            */
/* If the aset is used more than 1, it is cloned */
/* and the clone is modified                     */
/* asp : aset to add the aset to.                */
/* asp2 : aset to append.                        */
/* Return : the modified aset.                   */
/*          0 : error                            */
/* Link : asp  -1                                */
/*        asp2 -1                                */
/*        Return : 1                             */
/*************************************************/
e_aset *e_aset_append(e_aset *asp,e_aset *asp2)
{ e_aset *p;
  e_aset_item *aip2,*p1,*p2;
  e_name *np;
  void *valp;

#ifdef E_DEBUG
  e_debug("e_aset_append(asp=%x,asp2=%x)\n",asp,asp2);
#endif

  if(!asp) { e_error_null_pointer("e_aset_append_item(asp,-)"); goto error;}
  if(!asp2) { e_error_null_pointer("e_aset_append_item(-,asp2)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

  if(E_ASET_CARDINAL(asp2) == 0) /* asp2 is empty => return asp */
  { goto end;
  }

  if(E_ASET_CARDINAL(asp) == 0) /* asp is empty => return asp2 */
  { E_FREE(asp);
    return(asp2);
  }

  if(E_OBJ_REF(asp2) == 1) 
  {          /* Use asp2 as it is, no need to copy each item */
    p1 = E_ASET_LAST(asp);
    p2 = E_ASET_FIRST(asp2);
   
    E_ASET_ITEM_NEXT(p1) = p2;
    E_ASET_ITEM_PREV(p2) = p1;
    E_ASET_LAST(asp) = E_ASET_LAST(asp2);
    E_ASET_CARDINAL(asp) += E_ASET_CARDINAL(asp2);

    E_ASET_CARDINAL(asp2) = 0;
    E_ASET_FIRST(asp2) = 0;
    E_ASET_LAST(asp2) = 0;
    goto end;
  }

             /* Else : Append each element one by one */
  aip2 = E_ASET_FIRST(asp2);

  while(aip2)
  { if((np = E_ASET_ITEM_NAME(aip2))) E_REF(np);
    if((valp = E_ASET_ITEM_VAL(aip2))) E_REF(valp);

    E_NULL_ERROR(asp = e_aset_append_item(asp,np,valp));

    aip2 = E_ASET_ITEM_NEXT(aip2);
  }

end:
  E_FREE(asp2);
  return(asp);

error:
  if (asp) E_FREE(asp);  
  if (asp2) E_FREE(asp2);  
  return(0);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_fill(e_aset *asp,e_aset *asp2)
 *                     [-1,-1;+1]
 **/
/******************************************************************/
/* Fill in one aset with another one:                             */
/* if key in asp2 match key in asp => asp[key] = asp2[key]        */
/*   the value is overwritten.                                    */
/* Multiple keys in asp2 will overwrite themselves.               */
/* If a multiple key exists inasp, only the first one is modified */
/* key that have no match or element with no key are appended.    */
/* Item (key) that are write protected are left unchanged:        */
/*    no error and the pair key=val is not appended either.       */
/* If the asp is used more than 1, it is cloned and the clone     */
/* is modified                                                    */
/* TRICK : to ensure a set has unique keys fill in an empty aset: */
/*    asp = e_aset_fill(e_aset_new(0),asp);                       */
/******************************************************************/
/* asp : aset to fill.                                            */
/* asp2 : values to use.                                          */
/* Return : the modified aset.                                    */
/*          0 : error                                             */
/* Link : asp  -1                                                 */
/*        asp2 -1                                                 */
/*        Return : 1                                              */
/******************************************************************/
e_aset *e_aset_fill(e_aset *asp,e_aset *asp2)
{ e_aset *p;
  e_aset_item *aip2,*aip3;
  e_name *key;
  void *valp;

#ifdef E_DEBUG
  e_debug("e_aset_fill(asp=%x,asp2=%x)\n",asp,asp2);
#endif

  if(!asp) { e_error_null_pointer("e_aset_append_item(asp,-)"); goto error;}
  if(!asp2) { e_error_null_pointer("e_aset_append_item(-,asp2)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

             /* Append each element one by one */
  aip2 = E_ASET_FIRST(asp2);

  while(aip2)
  { if((valp = E_ASET_ITEM_VAL(aip2))) E_REF(valp);
    key = E_ASET_ITEM_NAME(aip2);

    if((key) && (aip3 = e_aset_get_item_by_key(asp,E_USE(key),1)))
    { /* Key exists in aip3 : overwrite it value if not write protected */
      if(!E_OBJ_READ_ONLY(aip3))
      { if(E_ASET_ITEM_VAL(aip3)) E_FREE(E_ASET_ITEM_VAL(aip3));
        E_ASET_ITEM_VAL(aip3) = valp;
      }
    }
    else
    { /* append the value */
      if(key) E_REF(key);
      E_NULL_ERROR(asp = e_aset_append_item(asp,key,valp));
    }

    aip2 = aip2->next;
  }

  E_FREE(asp2);
  return(asp);

error:
  if (asp) E_FREE(asp);  
  if (asp2) E_FREE(asp2);  
  return(0);
}

/************************ MODIFY ASET ****************************************/
/*****************************************************************************/
/**
 *  e_aset_item *e_aset_scan_by_key(e_aset *asp, e_name *key,int mode);
 *                                 [0,-1,.;0]   
 */
/********************************************************************/
/* Scan the aset by key and return an aset_item                     */
/* asp : set to scan.                                               */
/* key : key to search                                              */
/* mode : E_NEXT_RESET  : set cur item                              */
/*        E_NEXT_GET_NEXT : get next item matching the name         */
/*        E_NEXT_GET_CUR : return current item                      */
/*        E_NEXT_GET_PREV : get prev item matching the name         */
/* RETURN : the current aip                                         */
/*          0 :error or no more items                               */
/* LINK : key : -1                                                  */
/********************************************************************/
/* EXAMPLE                                                          */
/*  e_aset *asp; e_aset_item *aip; e_name *key;                     */
/*  e_aset_scan_by_key(asp,0,E_NEXT_RESET);                         */
/*  key = e_name_new("test");                                       */
/*  while(aip = e_aset_scan_by_key(asp,E_USE(key),E_NEXT_GET_NEXT)) */
/*  { USE OR MODIFY aip }                                           */
/********************************************************************/
e_aset_item *e_aset_scan_by_key(e_aset *asp, e_name *key,int mode)
{ e_aset_item *aip;

  if(!asp) { E_SAFE_FREE(key); return(0);}
  
  switch(mode)
  { case E_NEXT_RESET: 
                       E_ASET_CUR(asp) = 0;
                       E_SAFE_FREE(key);
                       return(0);
                       break;
      break;
    case E_NEXT_GET_NEXT: 
                       if(!(aip = E_ASET_CUR(asp))) aip = E_ASET_FIRST(asp);
                       while(aip && (E_ASET_ITEM_NAME(asp)!= key))
		       { aip = E_ASET_ITEM_NEXT(aip);
		       }
                       E_ASET_CUR(asp) = aip;
                       E_SAFE_FREE(key);
                       return(aip);
                       break;
    case E_NEXT_GET_CUR: 
                       E_SAFE_FREE(key);
                       return(E_ASET_CUR(asp));
                       break;
    case E_NEXT_GET_PREV: 
                       if(!(aip = E_ASET_CUR(asp))) aip = E_ASET_LAST(asp);
                       while(aip && (E_ASET_ITEM_NAME(asp)!= key))
		       { aip = E_ASET_ITEM_PREV(aip);
		       }
                       E_ASET_CUR(asp) = aip;
                       E_SAFE_FREE(key);
                       return(aip);
                       break;
  }
  return(0);
}

/*****************************************************************************/
/**
 * int e_aset_get_item_mode_by_key(e_aset *asp, e_name *key,int n); 
 *                                [0,-1,.;.]
 **/
/*****************************************************/
/* Get the protection mode of a given item.          */
/* aset : the aset to search                         */
/* key  : the item key                               */
/* n : nth key occurence to search for               */
/*     n > 0 search from the start. first is 1.      */
/*     n < 0 search from the end. last is -1         */
/*     0 : return NULL                               */
/* RETURN : the item mode                            */
/*          E_OBJ_NOT_FOUND : NOT found              */
/*          0 : error                                */
/* LINK : key : -1                                   */
/*****************************************************/
int e_aset_get_item_mode_by_key(e_aset *asp, e_name *key,int n)
{ e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_item_mode_by_key(asp=%x,key=%x,n=%d)\n",asp,key,n);
#endif

  if(!asp || !key || !n) { E_SAFE_FREE(key); return(0); }

  if((aip = e_aset_get_item_by_key(asp,key,n)))
  { return(E_OBJ_MODE(aip));
  }
  return(E_OBJ_NOT_FOUND);
}

/*****************************************************************************/
/**
 * int e_aset_get_item_mode_by_index(e_aset *asp,int index); 
 *                                  [0,.;.]
 **/
/*****************************************************/
/* Get the protection mode of a given item.          */
/* aset : the aset to search                         */
/* index  : the item index                           */
/* RETURN : the item mode                            */
/*          E_OBJ_NOT_FOUND : NOT found              */
/* LINK : key : -1                                   */
/*****************************************************/
int e_aset_get_item_mode_by_index(e_aset *asp,int index)
{ e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_item_mode_by_index(asp=%x,index=%d)\n",asp,index);
#endif

  if((aip = e_aset_get_item_by_index(asp,index)))
  { return(E_OBJ_MODE(aip));
  }
  return(E_OBJ_NOT_FOUND);
}

/*****************************************************************************/
/**
 * int e_aset_set_item_mode_by_key(e_aset *asp, e_name *key,
 *                                              int mode,int op,int n); 
 *                         [0,0,.,.,.;.]
 **/
/*****************************************************/
/* Set the protection mode of a given item           */
/* THE aset IS NOT CLONED!                           */
/* aset : the aset to modify                         */
/* key  : the item key                               */
/* mode : the oject mode                             */
/* op : E_SET_OP_RESET : item mode becomes mode.     */
/*      E_SET_OP_SET : set the bits that are on      */
/*      E_SET_OP_UNSET : unset the bits that are on. */
/*      by default use E_SET_OP_RESET                */
/* n : nth key occurence to search for               */
/*     n > 0 search from the start. first is 1.      */
/*     n < 0 search from the end. last is -1         */
/*     0 : return NULL                               */
/* RETURN : E_OBJ_OK : OK                            */
/*          E_OBJ_NOT_FOUND : item not found         */
/* LINK : key : -1                                   */
/*****************************************************/
int e_aset_set_item_mode_by_key(e_aset *asp, e_name *key,int mode,int op,int n)
{ e_aset_item *aip;
  int count;

#ifdef E_DEBUG
  e_debug("e_aset_set_item_mode_by_key(asp=%x,key=%x,mode=%x,op=%d)\n",
          asp,key,mode,op);
#endif

  if(!asp || !key || !n) { E_SAFE_FREE(key); return(0); }

  if(!n)
  { aip = E_ASET_FIRST(asp);
    while(aip)
    { if(E_ASET_ITEM_NAME(aip)==key)
      { switch(op)
        { case E_SET_OP_SET :   E_OBJ_MODE(aip) |= mode; break; 
          case E_SET_OP_UNSET : E_OBJ_MODE(aip) &= ~mode; break; 
          default :             E_OBJ_MODE(aip) = mode; break;
        }
        count++;
      }
      aip = E_ASET_ITEM_NEXT(aip);
    }
    E_FREE(key);

    if(count) return(E_OBJ_OK);
    return(E_OBJ_NOT_FOUND);
  }

  if((aip = e_aset_get_item_by_key(asp,key,n)))
  { switch(op)
    { case E_SET_OP_SET :   E_OBJ_MODE(aip) |= mode; break; 
      case E_SET_OP_UNSET : E_OBJ_MODE(aip) &= ~mode; break; 
      default :        E_OBJ_MODE(aip) = mode; break;
    }
   return(E_OBJ_OK);
  }
  return(E_OBJ_NOT_FOUND);
}

/*****************************************************************************/
/**
 * int e_aset_set_item_mode_by_index(e_aset *asp,int index,int mode,int op); 
 *                         [0,.,.,.;.]
 **/
/*****************************************************/
/* Set the protection mode of a given item           */
/* THE aset IS NOT CLONED!                           */
/* aset : the aset to modify                         */
/* index  : the item index position                  */
/* mode : the oject mode                             */
/* op : E_SET_OP_RESET : item mode becomes mode.     */
/*      E_SET_OP_SET : set the bits that are on      */
/*      E_SET_OP_UNSET : unset the bits that are on. */
/*      by default use E_SET_OP_RESET                */
/* RETURN : E_OBJ_OK : OK                            */
/*          E_OBJ_NOT_FOUND : item not found         */
/* LINK : no change                                  */
/*****************************************************/
int e_aset_set_item_mode_by_index(e_aset *asp,int index,int mode,int op)
{ e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_set_item_mode_by_index(asp=%x,index=%d,mode=%x,op=%d\n",
          asp,index,mode,op);
#endif

  if(!asp) return(0);

  if((aip = e_aset_get_item_by_index(asp,index)))
  { switch(op)
    { case E_SET_OP_SET :   E_OBJ_MODE(aip) |= mode; break; 
      case E_SET_OP_UNSET : E_OBJ_MODE(aip) &= ~mode; break; 
      default :        E_OBJ_MODE(aip) = mode; break;
    }
   return(E_OBJ_OK);
  }
  return(E_OBJ_NOT_FOUND);
}

/*****************************************************************************/
/**
 * int e_aset_set_item_value(e_aset_item *aip,void *val);  
 *                           [0,-1;.] 
 **/
/***********************************************************/
/* Set the value of the nth item an item matching the key. */
/* aip : aset item to modify                               */
/* val : new value for the item (can be NULL).             */ 
/* RETURN : E_OBJ_OK : OK                                  */
/*          E_OBJ_MODE_READ_ONLY : item is write protected */
/*          0 : ERROR                                      */
/* LINK : val : -1                                         */
/***********************************************************/
int e_aset_set_item_value(e_aset_item *aip, void *val)
{ 
#ifdef E_DEBUG
  e_debug("e_aset_set_item_value(aip=%x,val=%x)\n",asp,val);
#endif

  if(!aip) return(0);

  if(E_OBJ_READ_ONLY(aip))
  { E_SAFE_FREE(val);
    return(E_OBJ_MODE_READ_ONLY);
  }

  E_SAFE_FREE(E_ASET_ITEM_VAL(aip));
  E_ASET_ITEM_VAL(aip) = val;
  return(E_OBJ_OK);
}

/*****************************************************************************/
/**
 * int e_aset_set_value_by_key(e_aset *asp, e_name *key, void *val,int n);  
 *                   [0,-1,-1;.] 
 **/
/***********************************************************/
/* Set the value of the nth item an item matching the key. */
/* asp : aset to modify                                    */
/* key : key for the item to modify                        */
/* val : new value for the item (can be NULL).             */ 
/* n : the n_th key (0 : ALL, -1 count from the end )      */     
/* RETURN : E_OBJ_OK : OK                                  */
/*          E_OBJ_NOT_FOUND : item not found               */
/*          E_OBJ_MODE_READ_ONLY : item is write protected */
/*          0 : error                                      */
/* LINK : key : -1                                         */
/*        val : -1                                         */
/***********************************************************/
int e_aset_set_value_by_key(e_aset *asp, e_name *key, void *val,int n)
{ e_aset_item *aip;
  int count;

#ifdef E_DEBUG
  e_debug("e_aset_set_val_by_key(asp=%x,key=%x,val=%x,n=%d)\n",asp,key,val,n);
#endif

  if(!asp || !key) return(0);

  count = 1;
  if(n == 0)
  { aip = E_ASET_FIRST(asp);
    while(aip)
    { if(E_ASET_ITEM_NAME(aip)==key)
      { if(val) E_REF(val);
        e_aset_set_item_value(aip,val);
        count++;
      }
      aip = E_ASET_ITEM_NEXT(aip);
    }
    E_SAFE_FREE(val);
    E_FREE(key);

    if(count) return(E_OBJ_OK);
    return(E_OBJ_NOT_FOUND);
  }

  if(n >= 0)
  { if((aip = e_aset_get_item_by_key(asp,E_USE(key),n)))
    { E_FREE(key);
      return(e_aset_set_item_value(aip,val));
    }
  }

  E_SAFE_FREE(val);
  E_FREE(key);
  return(E_OBJ_NOT_FOUND);
}

/*****************************************************************************/
/**
 * int e_aset_set_value_by_index(e_aset *asp,int index, void *val);  
 *                   [0,.,-1;.] 
 **/
/***********************************************************/
/* Set the value of an item by key                         */
/* asp : aset to modify                                    */
/* index : index for the item to modify                    */
/*    (first is 0, last is -1)                             */
/* val : new value for the item (can be NULL).             */      
/* RETURN : E_OBJ_OK : OK                                  */
/*          E_OBJ_NOT_FOUND : item not found               */
/*          E_OBJ_MODE_READ_ONLY : item is write protected */
/* LINK : val : -1                                         */
/***********************************************************/
int e_aset_set_value_by_index(e_aset *asp,int index, void *val)
{ e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_set_val_by_index(asp=%x,index=%d,val=%x)\n",asp,index,val);
#endif

  if((aip = e_aset_get_item_by_index(asp,index)))
  { if(E_OBJ_READ_ONLY(aip))
    { E_SAFE_FREE(val);
      return(E_OBJ_MODE_READ_ONLY);
    }
    E_SAFE_FREE(E_ASET_ITEM_VAL(aip));
    E_ASET_ITEM_VAL(aip) = val;
    return(E_OBJ_OK);
  }
  E_SAFE_FREE(val);
  return(E_OBJ_NOT_FOUND);
}

/*****************************************************************************/
/**
 * e_aset *e_aset_apply_fct(e_aset *asp,void *p,int i,
 *                          e_obj(* fct)(e_obj *op,void *p,int i));
 *                          [-1,.,.,.;+1] 
 **/
/******************************************************************/
/* Apply the function fct to every item value of the aset.        */
/* If the function returns a non null value, the item             */
/* value is replaced by the returned object.                      */
/* The aset returned is build from scratch.                       */
/* asp : aset to apply the fct too.                               */
/* p : a pointer parameter from the applied function.             */
/* i : a integer parameter from the applied function.             */
/* fct : function to apply.                                       */
/*       e_obj : value to modify. Can be null. LINK : -1          */
/*       p and i : pointer and integer parameter.                 */
/*       RETURN : The modified value : LINK +1                    */
/*                0 : ignore that element                         */
/* RETURN : the modified aset (always a new aset!)                */
/* LINK : asp : -1                                                */
/*        RETURN : +1                                             */
/******************************************************************/
e_aset *e_aset_apply_fct(e_aset *asp,void *p,int i,
                   e_obj *(* fct)(e_obj *op,void *p,int i))
{ e_aset_item *aip;
  e_aset *asp2=0;
  e_name *key;
  void *obj;

#ifdef E_DEBUG
  e_debug("e_aset_apply_fct(asp=%x,p=%x,i=%d,fct=%x)\n",asp,p,i,fct);
#endif

  if(!asp || !fct) goto error;

  E_NULL_ERROR(asp2 = e_aset_new(0));

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if((obj = E_ASET_ITEM_VAL(aip))) E_USE(obj);
  if((obj = fct(obj,p,i)))
    { if((key = E_ASET_ITEM_NAME(aip))) E_USE(key);
       E_NULL_ERROR(asp2 = e_aset_append_item(asp2,key,obj));
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  E_FREE(asp);
  return(asp2);

error:
  E_SAFE_FREE(asp2);
  return(0);
}


/************************ EXTRACT FROM ASET ********************************/

/*****************************************************************************/
/**
 *  int e_aset_get_key_index(e_aset *asp, e_name *np,int start)
 *                          [0,-1,.;.]
 **/
/*********************************************************************/
/* Return the index of key in the aset staring at position start     */
/* The first element has index 0                                     */
/* Example :                                                         */    
/*  i = 0;                                                           */
/*  while((i = e_aset_get_key_index(asp, e_name_new("test"),i)) > 0) */
/*  { printf("index = %d\n",i++);                                    */
/*  }                                                                */
/* asp : aset to scan.                                               */
/* np : key name                                                     */
/* start : index of the first element to check                       */
/* Return : index position of the key in the set                     */
/*          -1 : key not found                                       */
/*          -2 : error                                               */
/* Link : asp  0                                                     */
/*        np  -1                                                     */
/*********************************************************************/
int e_aset_get_key_index(e_aset *asp, e_name *np,int start)
{ int i; 
  e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_key_index(asp=%x,np=%x,start=%d)\n",asp,np,start);
#endif

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,"e_aset_get_key_index(asp,-,-)"));
  E_NULL_ERROR(e_check_type(np,E_TYPE_NAME,"e_aset_get_key_index(-,np,-)"));

  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* Move to start */
  while((i < start) && aip)
  { aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  
  while(aip)
  { if(E_ASET_ITEM_NAME(aip) == np) 
    { E_FREE(np);
      return(i);
    }
    aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  E_FREE(np);
                       /* Key not found */
  return(-1);

error:
  if(np) E_FREE(np);
  return(-2);  
}

/*****************************************************************************/
/**
 *  int e_aset_get_value_index(e_aset *asp,void *obj,int start)
 *                            [0,-1,.;.]
 **/
/*************************************************************************/
/* Return the index of the element obj in the aset staring at position   */
/* start. Return -1 if obj does not belong to the set.                   */
/* The first element has index 0                                         */
/* Example :                                                             */    
/*  i = 0;                                                               */
/*  while((i = e_aset_get_value_index(asp,val,i)) > 0)                   */
/*  { printf("index = %d\n",i++);                                        */
/*  }                                                                    */
/* asp : aset to scan.                                                   */
/* obj : object to find                                                  */
/* start : index of the first element to check                           */
/* Return : index position of the key in the set                         */
/*          -1 : object not found                                        */
/*          -2 : error                                                   */
/* Link : asp  0                                                         */
/*        obj  -1                                                        */
/*************************************************************************/
int e_aset_get_value_index(e_aset *asp,void *obj,int start)
{ int i; 
  e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_value_index(asp=%x,obj=%x,start=%d)\n",asp,obj,start);
#endif

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,
                              "e_aset_get_value_index(asp,-,-)"));
  E_NULL_ERROR(obj);

  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* Move to start */
  while((i < start) && aip)
  { aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  
  while(aip)
  { if(E_EQUALS(E_ASET_ITEM_VAL(aip),obj)) 
    { E_FREE(obj);
      return(i);
    }
    aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  E_FREE(obj);
                       /* Key not found */
  return(-1);

error:
  if(obj) E_FREE(obj);
  return(-2);  
}

/*****************************************************************************/
/**
 *  int e_aset_get_item_index(e_aset *asp, e_name *np,void *val,int start)
 *                        [0,-1,-1,.;.]
 **/
/*********************************************************************/
/* Return the index of an item (key,val) pair in the aset staring at */
/* the position start.                                               */
/* The first element has index 0                                     */
/* Example :                                                         */    
/*  i = 0;                                                           */
/*  while((i = e_aset_get_item_index(asp,key,val,i)) > 0)            */
/*  { printf("index = %d\n",i++);                                    */
/*  }                                                                */
/* asp : aset to scan.                                               */
/* np : key name to match                                            */
/* val : value to match (can be NULL to match key with no value)     */
/* start : index of the first element to check                       */
/* Return : index position of the key in the set                     */
/*          -1 : item not found                                      */
/*          -2 : error                                               */
/* Link : asp  0                                                     */
/*        np  -1                                                     */
/*        val -1                                                     */
/*********************************************************************/
int e_aset_get_item_index(e_aset *asp, e_name *np,void *val,int start)
{ int i,ret=-1; 
  e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_item_index(asp=%x,np=%x,val=%x,start=%d)\n",
          asp,np,val,start);
#endif

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,"e_aset_get_key_index(asp,-,-)"));
  E_NULL_ERROR(e_check_type(np,E_TYPE_NAME,"e_aset_get_key_index(-,np,-)"));
  
  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* Move to start */
  while((i < start) && aip)
  { aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  
  while(aip)
  { if(E_ASET_ITEM_NAME(aip) == np) 
    { if (val == E_ASET_ITEM_VAL(aip)) { ret = i; goto end;}
      if (!val || E_EQUALS(E_ASET_ITEM_VAL(aip),val)) { ret = i; goto end;}
    }

    aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }

end:
  E_FREE(np);
  E_FREE(val);
                       /* Key not found */
  return(ret);

error:
  if(np) E_FREE(np);
  return(-2);  
}

/*****************************************************************************/
/**
 *  void *e_aset_get_value_by_key(e_aset *asp, e_name *np,int start)
 *                               [0,-1,.;+1]
 **/
/***********************************************************************/
/* Return the value associated with the key staring at position start. */
/* The first aset item has index 0.                                    */
/* Example :                                                           */    
/*  obj = e_aset_get_key_value(asp,e_name_new("test"),0);              */
/* asp : aset to scan.                                                 */
/* np : key name                                                       */
/* start : index of the first element to check                         */
/* Return : value associated with the key                              */
/*          0 : key not found or error                                 */
/* Link : asp  0                                                       */
/*        np  -1                                                       */
/*        Return : +1                                                  */
/***********************************************************************/
void *e_aset_get_value_by_key(e_aset *asp, e_name *np,int start)
{ int i; 
  e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_value_by_key(asp=%x,np=%x,start=%d)\n",asp,np,start);
#endif

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,
                              "e_aset_get_value_by_key(asp,-,-)"));
  E_NULL_ERROR(e_check_type(np,E_TYPE_NAME,
                              "e_aset_get_value_by_key(-,np,-)"));

  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* Move to start */
  while((i < start) && aip)
  { aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  
  while(aip)
  { if(E_ASET_ITEM_NAME(aip) == np) 
    { E_FREE(np);
      if(E_ASET_ITEM_VAL(aip)) E_REF(E_ASET_ITEM_VAL(aip));
      return(E_ASET_ITEM_VAL(aip));
    }
    aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  E_FREE(np);
                       /* Key not found */
  return(0);

error:
  if(np) E_FREE(np);
  return(0);  
}

/*****************************************************************************/
/**
 *  void *e_aset_get_value_by_index(e_aset *asp, int index)
 *                                 [-0,.;+1]
 **/
/***********************************************************************/
/* Get an aset element value using its index position in the aset.     */
/* The first element has index 0.                                      */
/* Example :                                                           */    
/*  obj = e_aset_get_value_by_index(asp,0);                            */
/* asp : aset to scan.                                                 */
/* index : position of object, if negative, count from the end.        */
/*         0 : first element; -1 : last element.                       */
/*         if the index is invalid, it returns NULL but not error      */
/*         are generated.                                              */
/* RETURN : value associated with the key                              */
/*          0 : object not found                                       */
/*          0 : asp is not an aset.                                    */
/* LINK : asp  0                                                       */
/*        Return : +1                                                  */
/***********************************************************************/
void *e_aset_get_value_by_index(e_aset *asp, int index)
{ int i,j; 
  e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_value_by_index(asp=%x,index=%d)\n",asp,index);
#endif

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,
                         "e_aset_get_value_by_index(asp,-)"));

  if(index < 0) i = E_ASET_CARDINAL(asp)+index;
  else i = index;

  if ((i < 0) || (i >= E_ASET_CARDINAL(asp)))
  {                          /* The aset member does not exist */
    return(0);
  }

  if (i > (E_ASET_CARDINAL(asp)/2))
  {                          /* second half of aset: start at the end */
    j = E_ASET_CARDINAL(asp)-1;
    aip = E_ASET_LAST(asp);
    while((j > i) && aip)
    { aip = E_ASET_ITEM_PREV(aip);
      j--;
    }
  }
  else
  {                         /* first half of aset */
    j = 0;
    aip = E_ASET_FIRST(asp);

    while((j < i) && aip)
    { aip = E_ASET_ITEM_NEXT(aip);
      j++;
    }
  }

  if(!aip)                 /* should never happen */
  { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
        "Aset inconsistency : fewer element than cardinal",(unsigned int)asp);
    return(0);       
  }

  if(E_ASET_ITEM_VAL(aip)) E_REF(E_ASET_ITEM_VAL(aip));
  return(E_ASET_ITEM_VAL(aip));

error:
  return(0);  
}

/*****************************************************************************/
/**
 *  e_aset_item *e_aset_get_item_by_key(e_aset *asp, e_name *np,int n)
 *                                     [0,-1,.;0]
 **/
/******************************************************/
/* Return a pointer to the aset_item for nthe key np. */
/* To use with care.                                  */
/* asp : aset to scan.                                */
/* np : name for the key to search.                   */
/* n : nth key occurence to search for                */
/*     n > 0 search from the start. first is 1.       */
/*     n < 0 search from the end. last is -1          */
/*     0 : return NULL                                */
/* RETURN : aset_item for key np.                     */
/*          0 : error or item not found               */
/* LINK asp : 0                                       */
/*      np  : -1                                      */
/*      Return  : 0                                   */
/******************************************************/
e_aset_item *e_aset_get_item_by_key(e_aset *asp, e_name *np,int n)
{ e_aset_item *aip;
  int count;

#ifdef E_DEBUG
  e_debug("e_aset_get_item_by_key(asp=%x,np=%x,n=%d)\n",asp,np,n);
#endif

  if(!n) { E_SAFE_FREE(np); return(0); }

  count = 1;
  aip = E_ASET_FIRST(asp);

  if(n > 0)
  { while(aip)
    { if(E_ASET_ITEM_NAME(aip) == np)
      { if (count == n)
        { E_SAFE_FREE(np);
          return(aip);
        }
        count++;
      }
      aip = E_ASET_ITEM_NEXT(aip);
    }
  }
  else /* n < 0 */
  { while(aip)
    { if(E_ASET_ITEM_NAME(aip) == np)
      { if (count == -n)
        { E_SAFE_FREE(np);
          return(aip);
        }
        count++;
      }
      aip = E_ASET_ITEM_NEXT(aip);
    }
  }

  E_SAFE_FREE(np);
  return(0);
}

/*****************************************************************************/
/**
 *  e_aset_item *e_aset_get_item_by_value(e_aset *asp, void *val)
 *                                       [0,-1;0]
 **/
/*********************************************************/
/* Return a pointer to the aset_item for the object val. */
/* To use with care.                                     */
/* asp : aset to scan.                                   */
/* val : the object to search.                           */
/* RETURN : aset_item for val.                           */
/*          0 : error or item not found                  */
/* LINK asp : 0                                          */
/*      val : -1                                         */
/*      Return  : 0                                      */
/*********************************************************/
e_aset_item *e_aset_get_item_by_value(e_aset *asp, void *val)
{ e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_item_by_value(asp=%x,val=%x)\n",asp,val);
#endif

  if(!val || !asp) return(0);

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if((val == E_ASET_ITEM_VAL(aip)) || 
       (E_ASET_ITEM_VAL(aip) && E_EQUALS(val,E_ASET_ITEM_VAL(aip)))) 
    { E_FREE(val);
      return(aip);
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  E_FREE(val);

  return(0);
}

/****************************************************************************/
/**
 *  e_aset_item *e_aset_get_item_by_index(e_aset *asp,int index)
 *                                       [0,.;0]
 **/
/**********************************************************/
/* Return a pointer to the aset_item for the given index. */
/* To use with care.                                      */
/* asp : aset to scan.                                    */
/* index : index position of the item in the aset.        */
/* RETURN : aset_item for the index.                      */
/*          0 : error or item not found                   */
/* LINK asp : 0                                           */
/*      RETURN : 0                                        */
/**********************************************************/
e_aset_item *e_aset_get_item_by_index(e_aset *asp,int index)
{ e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_get_item_by_index(asp=%x,index=%d)\n",asp,index);
#endif  

  if (index < 0) index += E_ASET_CARDINAL(asp);
  if ((index < 0) || (index >= E_ASET_CARDINAL(asp))) return(0);

  aip = E_ASET_FIRST(asp);
  while(aip && index)
  { aip = E_ASET_ITEM_NEXT(aip);
    index--;
  }
  return(aip);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_get_keys(e_aset *asp)
 *                         [0;+1]
 **/
/*********************************************************/
/* List the aset keys into an aset                       */
/* The returned set is an aset of e_string               */
/* asp : aset to scan.                                   */
/* RETURN : aset with all the keys.                      */
/*          0 : error                                    */
/* Link asp : 0                                          */
/*      Return  : +1                                     */
/*********************************************************/
e_aset *e_aset_get_keys(e_aset *asp)
{ e_aset_item *aip;
  e_aset *asp2=0;
  e_name *key;

#ifdef E_DEBUG
  e_debug("e_aset_get_keys(asp=%x)\n",asp);
#endif

  if(!asp) return(0);

  E_NULL_ERROR(asp2 = e_aset_new(0));

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if((key=E_ASET_ITEM_NAME(aip)))
    { E_NULL_ERROR(asp2 = e_aset_append_item(asp2,0,
                                    e_string_new(0,E_NAME_NAME(key),0)));
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  return(asp2);

error:
  E_SAFE_FREE(asp2);
  return(0);
}


/*****************************************************************************/
/**
 *  e_aset *e_aset_anonymous_subset(e_aset *asp)
 *                         [0;+1]
 **/
/*********************************************************/
/* Return the list of anonymous elements                 */
/* asp : aset to scan.                                   */
/* RETURN : aset of anonymous elements                   */
/*          0 : error                                    */
/* Link asp : 0                                          */
/*      Return  : +1                                     */
/*********************************************************/
e_aset *e_aset_anonymous_subset(e_aset *asp)
{ e_aset_item *aip;
  e_aset *asp2=0;
  void *obj;

#ifdef E_DEBUG
  e_debug("e_aset_anonymous_subset(asp=%x)\n",asp);
#endif

  if(!asp) return(0);

  E_NULL_ERROR(asp2 = e_aset_new(0));

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if(!E_ASET_ITEM_NAME(aip))
    { if((obj = E_ASET_ITEM_VAL(aip)))
      { E_USE(obj);
        E_NULL_ERROR(asp2 = e_aset_append_item(asp2,0,obj));
      }
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  return(asp2);

error:
  E_SAFE_FREE(asp2);
  return(0);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_subset(e_aset *asp,int from,int to)
 *                         [0,.,.;+1]
 **/
/*********************************************************/
/* Return the subset asp[from .. to]                     */
/* asp : aset to scan.                                   */
/* from : first index.                                   */
/* to : last index.                                      */
/* Indices are interporeted as follow:                   */
/*   first index is 0                                    */
/*   index < 0 : count from last : (last is -1)          */
/*   if outside the bound : reajust them                 */
/* RETURN : subset                                       */
/*          0 : error                                    */
/* Link asp : 0                                          */
/*      Return  : +1                                     */
/*********************************************************/
e_aset *e_aset_subset(e_aset *asp,int from,int to)
{ e_aset_item *aip;
  e_aset *asp2=0;
  e_name *key;
  void *obj;
  int n;

#ifdef E_DEBUG
  e_debug("e_aset_subset(asp=%x,from=%d,to=%d)\n",asp,from,to);
#endif

  if(!asp) return(0);

  if(from < 0) from = E_ASET_CARDINAL(asp)+from;
  if(from >= E_ASET_CARDINAL(asp)) from = E_ASET_CARDINAL(asp)-1;
  if(from < 0) from = 0;

  if(to < 0)   to = E_ASET_CARDINAL(asp)+to;
  if(to >= E_ASET_CARDINAL(asp)) to = E_ASET_CARDINAL(asp)-1;
  if(to < 0) to = 0;

  E_NULL_ERROR(asp2 = e_aset_new(0));

  aip = E_ASET_FIRST(asp);
  n = 0;

  while(aip)
  { if( (from <= n) && (n <= to))
    { if((key=E_ASET_ITEM_NAME(aip))) E_USE(key);
      if((obj=E_ASET_ITEM_VAL(aip)))  E_USE(obj);

      E_NULL_ERROR(asp2 = e_aset_append_item(asp2,key,obj));
    }
    n++;
    if(n > to) aip = 0;
    else aip = E_ASET_ITEM_NEXT(aip);
  }
  return(asp2);

error:
  E_SAFE_FREE(asp2);
  return(0);
}

/*********************************************************/
/**
 *  e_aset *e_aset_subset_by_key_match(e_aset *asp,e_string *match);
 *                                     [0,-1;+1]
 **/
/*********************************************************/
/* Return the list of named elements                     */
/* asp : aset to scan.                                   */
/* RETRUN : aset of anonymous elements                   */
/*          0 : error                                    */
/* Link asp : 0                                          */
/*      Return  : +1                                     */
/*********************************************************/
e_aset *e_aset_subset_by_key_match(e_aset *asp,e_string *match)
{ e_aset_item *aip;
  e_aset *asp2=0;
  e_name *key;
  void *obj;
  char *pattern;

#ifdef E_DEBUG
  e_debug("e_aset_subset_by_key_match(asp=%x,match=%x)\n",asp,match);
#endif

  if(!asp || !match) goto error;

  E_NULL_ERROR(asp2 = e_aset_new(0));

  aip = E_ASET_FIRST(asp);
  pattern = E_STRING_S(match);

  while(aip)
  { if((key = E_ASET_ITEM_NAME(aip)) && 
       e_tools_unix_match(pattern,E_NAME_NAME(key)))
    { if((obj = E_ASET_ITEM_VAL(aip))) E_USE(obj);
       E_NULL_ERROR(asp2 = e_aset_append_item(asp2,E_USE(key),obj));
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  return(asp2);

error:
  E_SAFE_FREE(asp2);
  E_SAFE_FREE(match);
  return(0);
}

/*********************************************************/
/**
 *  e_aset *e_aset_subset_by_test(e_aset *asp,void *val,
 *                                int (*test)(e_obj *op,void *val))
 *                               [0,0,0;+1]
 **/
/*********************************************************/
/* Return the list of named elements                     */
/* asp : aset to scan.                                   */
/* RETRUN : aset of anonymous elements                   */
/*          0 : error                                    */
/* Link asp : 0                                          */
/*      Return  : +1                                     */
/*********************************************************/
e_aset *e_aset_subset_by_test(e_aset *asp,void *val,
                              int (*test)(e_obj *op,void *val))
{ e_aset_item *aip;
  e_aset *asp2=0;
  e_name *key;
  void *obj;

#ifdef E_DEBUG
  e_debug("e_aset_subset_by_test(asp=%x,cal=%x,test=%x)\n",asp,val,test);
#endif

  if(!asp || !val || !test) goto error;

  E_NULL_ERROR(asp2 = e_aset_new(0));

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if((obj = E_ASET_ITEM_VAL(aip)) && test(obj,val))
    { if((key = E_ASET_ITEM_NAME(aip))) E_USE(key);
      E_NULL_ERROR(asp2 = e_aset_append_item(asp2,key,E_USE(obj)));
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  return(asp2);

error:
  E_SAFE_FREE(asp2);
  return(0);
}

/*********************************************************/
/**
 *  char *e_aset_get_string_pointer_by_key(e_aset *asp,e_name *np)
 *                               [0,0;-]
 **/
/*********************************************************/
/* Return the string pointer of the value of the item    */
/* matching the key in the aset.                         */
/* The value must be a string or a name                  */
/* asp : aset to scan.                                   */
/* np : key to match.                                    */
/* WARNING : the pointer is only valid for as long as    */
/*           the aset item exists!                       */
/* RETURN : pointer to string value                      */
/*          0 : error, no match, value is not a string   */
/* Link asp : 0                                          */
/*       np : 0                                          */
/*********************************************************/
char *e_aset_get_string_pointer_by_key(e_aset *asp,e_name *np)
{ e_aset_item *aip;
  e_string *val;

#ifdef E_DEBUG
  e_debug("e_aset_get_string_pointer_by_key(asp=%x,np=%x)\n",asp,name);
#endif

  if(!asp || !np) goto error;

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if((E_ASET_ITEM_NAME(aip) == np) && (val=E_ASET_ITEM_VAL(aip)))
    { if (E_IS_STRING(val))
      { return(E_STRING_S(val));
      }
      else if (E_IS_NAME(val))
      { return(E_NAME_NAME(val));
      }
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }

error:
  return(0);
}

/*********************************************************/
/**
 *  void *e_aset_get_number_pointer_by_key(e_aset *asp,e_name *np,int type)
 *                               [0,0,-;-]
 **/
/*************************************************************/
/* Return the number pointer of the value of the item        */
/* matching the key in the aset.                             */
/* The value must be a number                                */
/* asp : aset to scan.                                       */
/* np : key to match.                                        */
/* type : integer type (E_NUM_TYPE_CHAR ... E_NUM_TYPE_COMP) */
/* WARNING : the pointer is only valid for as long as        */
/*           the aset item exists!                           */
/* RETURN : pointer to a number (*char, *int ... *comp)      */
/*          0 : error, no match, value is not a string       */
/* Link asp : 0                                              */
/*       np : 0                                              */
/*************************************************************/
void *e_aset_get_number_pointer_by_key(e_aset *asp,e_name *np,int type)
{ e_aset_item *aip;
  e_string *val;

#ifdef E_DEBUG
  e_debug("e_aset_get_number_pointer_by_key(asp=%x,np=%x,type=%d)\n",
           asp,name,type);
#endif

  if(!asp || !np) goto error;

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if((E_ASET_ITEM_NAME(aip) == np) && (val=E_ASET_ITEM_VAL(aip))
       && E_IS_NUMBER(val) && (type == E_NUMBER_TYPE(val)) )
    { switch(type)
      { case E_NUM_TYPE_CHAR:
	     return(&E_NUMBER_VAL_CHAR(val));
   	     break;
        case E_NUM_TYPE_INT:
	     return(&E_NUMBER_VAL_INT(val));
   	     break;
        case E_NUM_TYPE_UINT:
	     return(&E_NUMBER_VAL_UINT(val));
   	     break;
        case E_NUM_TYPE_REAL:
	     return(&E_NUMBER_VAL_REAL(val));
   	     break;
        case E_NUM_TYPE_COMP:
	     return(&E_NUMBER_VAL_COMP(val));
   	     break;
      }
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }

error:
  return(0);
}

/***************************** MODIFY ASET **********************************/
/*****************************************************************************/
/**
 *  e_aset *e_aset_sort_by_key(e_aset *asp,e_flag mode)
 *                            [-1,.;+1]
 **/
/**************************************************************************/
/* Sort e_aset by key                                                     */
/* Clone the aset recursively and sort the clone.                         */
/* Anonymous elements are always pushed at the end                        */
/* mode : E_MODE_INCREASING or E_MODE_DECREASING : sorting order          */
/*        E_MODE_NOCLONE ; don't clone the original, asp is not pointed to*/
/* Return : sorted e_aset                                                 */
/*          0 : Error                                                     */
/* LINK : asp : -1                                                        */
/*        RETURN : +1                                                     */
/**************************************************************************/
e_aset *e_aset_sort_by_key(e_aset *asp,e_flag mode)
{ e_aset *new;
  e_aset_item *item,*next;
  e_name *n1,*n2;
  void *tmp;
  int swap,order,test;

#ifdef E_DEBUG
  e_debug("e_aset_sort_by_key(asp=%x,mode=%x)\n",asp,mode);
#endif

  if(!asp || (E_ASET_CARDINAL(asp) < 2)) return(asp);

  if(!(mode & E_MODE_NOCLONE))
  { E_NULL_ERROR(new = e_aset_clone(asp,E_CLONE_RECURSIVE));
    E_FREE(asp);
    asp = new;
  }

  if(mode & E_MODE_DECREASING) order = -1;
  else order = 1;

  swap = 1;
  while(swap) /* Sort until no more swaps */
  { swap = 0;
    item = E_ASET_FIRST(asp);
    next = E_ASET_ITEM_NEXT(item);

    while(next)
    { /* Compare two succesive elements */

      n1 = E_ASET_ITEM_NAME(item);
      n2 = E_ASET_ITEM_NAME(next);
      if(n1 && n2)
      { test = strcmp(E_NAME_NAME(n1),E_NAME_NAME(n2))*order;
      }
      else if (!n2) test = -1;
      else test = 1;

      if (test > 0)
      { /* Swap them : swap the KEY and the value */

        tmp = E_ASET_ITEM_NAME(item);
        E_ASET_ITEM_NAME(item) = E_ASET_ITEM_NAME(next);
        E_ASET_ITEM_NAME(next) = tmp;

        tmp = E_ASET_ITEM_VAL(item);
        E_ASET_ITEM_VAL(item) = E_ASET_ITEM_VAL(next);
        E_ASET_ITEM_VAL(next) = tmp;

        swap += 1;
      }

      item = next;
      next = E_ASET_ITEM_NEXT(item);

    }
  }

  return(asp);

error:
  E_FREE(asp);
  return(0);
}

/***************************** MODIFY ASET **********************************/
/*****************************************************************************/
/**
 *  e_aset *e_aset_sort_by_value(e_aset *asp,e_flag mode,
 *                                 int (* testf)(void *,void *))
 *                              [-1,.,.;+1]
 **/
/**************************************************************************/
/* Sort e_aset by value                                                   */
/* Clone the aset recursively and sort the clone.                         */
/* asp : aset to sort                                                     */
/* mode : E_MODE_INCREASING or E_MODE_DECREASING : sorting order          */
/*        E_MODE_NOCLONE ; don't clone the original, asp is not pointed to*/
/*        mode is passed to the test function.                            */
/* test(val1,val2,mode) : function used to perform the test [0,0,.]       */
/*      return : (val1 > val2) : 1;  (val1==val2) : 0; otherwise -1       */
/* Return : sorted e_aset                                                 */
/*          0 : Error                                                     */
/* LINK : asp : -1                                                        */
/*        RETURN : +1                                                     */
/**************************************************************************/
e_aset *e_aset_sort_by_value(e_aset *asp,e_flag mode,
                             int (* testf)(void *,void *,int))
{ e_aset *new;
  e_aset_item *item,*next;
  void *tmp;
  int swap;

#ifdef E_DEBUG
  e_debug("e_aset_sort_by_value(asp=%x,mode=%x,testf=%x)\n",asp,mode,testf);
#endif

  if(!asp || (E_ASET_CARDINAL(asp) < 2)) return(asp);

  if(!(mode & E_MODE_NOCLONE))
  { E_NULL_ERROR(new = e_aset_clone(asp,E_CLONE_RECURSIVE));
    E_FREE(asp);
    asp = new;
  }

  swap = 1;
  while(swap) /* Sort until no more swaps */
  { swap = 0;
    item = E_ASET_FIRST(asp);
    next = E_ASET_ITEM_NEXT(item);

    while(next)
    { /* Compare two succesive elements */
      if (testf(E_ASET_ITEM_VAL(item),E_ASET_ITEM_VAL(next),mode) > 0)
      { /* Swap them : swap the KEY and the value */

        tmp = E_ASET_ITEM_NAME(item);
        E_ASET_ITEM_NAME(item) = E_ASET_ITEM_NAME(next);
        E_ASET_ITEM_NAME(next) = tmp;

        tmp = E_ASET_ITEM_VAL(item);
        E_ASET_ITEM_VAL(item) = E_ASET_ITEM_VAL(next);
        E_ASET_ITEM_VAL(next) = tmp;

        swap += 1;
      }

      item = next;
      next = E_ASET_ITEM_NEXT(item);

    }
  }

  return(asp);

error:
  E_FREE(asp);
  return(0);
}

/*****************************************************************************/
/**
 *  int e_aset_remove_item_by_index(e_aset *asp,int index)
 *                                 [0,.;.]
 **/
/**************************************************/ 
/* remove item from the aset by index             */
/* asp : aset to modify                           */
/* index : item position. First item has index=0  */
/* The aset is not cloned!                        */
/* No link changed                                */
/* RETURN : 1 : OK                                */
/*          0 : error                             */
/**************************************************/ 
int e_aset_remove_item_by_index(e_aset *asp,int index)
{ int i;
  e_aset_item *aip,*prev,*next;

#ifdef E_DEBUG
  e_debug("e_aset_remove_item_by_index(asp=%x,index=%d)\n",asp,index);
#endif

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,
                            "e_aset_remove_item_bu_index(asp,-)"));

  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* search item to remove */
  while((i < index) && aip)
  { aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }

  if(!aip) goto error;  

  prev = E_ASET_ITEM_PREV(aip);
  next = E_ASET_ITEM_NEXT(aip);

  if(prev)
  { E_ASET_ITEM_NEXT(prev) = next;
  }
  else
  { E_ASET_FIRST(asp) = next; 
  }

  if(next)
  { E_ASET_ITEM_PREV(next) = prev;
  }
  else
  { E_ASET_LAST(asp) = prev; 
  }

  E_ASET_CARDINAL(asp)--;

  return(1);

error:
  return(0);  
}

/*****************************************************************************/
/**
 *  int e_aset_remove_cur_item(e_aset *asp)
 *                            [0;.]
 **/
/**************************************************/ 
/* remove the current item from the aset          */
/* asp : aset to modify                           */
/* The aset is not cloned!                        */
/* No link changed                                */
/* RETURN : 1 : OK                                */
/*          0 : error                             */
/**************************************************/ 
int e_aset_remove_cur_item(e_aset *asp)
{ e_aset_item *aip,*prev,*next;

#ifdef E_DEBUG
  e_debug("e_aset_remove_cur_item(asp=%x)\n",asp);
#endif

  E_NULL_ERROR(asp);

  aip = E_ASET_CUR(asp);

  if(!aip) return(1); /* nothing to remove */  

  prev = E_ASET_ITEM_PREV(aip);
  next = E_ASET_ITEM_NEXT(aip);
  E_ASET_CUR(asp) = next;

  if(prev)
  { E_ASET_ITEM_NEXT(prev) = next;
  }
  else
  { E_ASET_FIRST(asp) = next; 
  }

  if(next)
  { E_ASET_ITEM_PREV(next) = prev;
  }
  else
  { E_ASET_LAST(asp) = prev; 
  }

  E_ASET_CARDINAL(asp)--;

  return(1);

error:
  return(0);  
}

/*****************************************************************************/
/**
 *  int e_aset_remove_item_by_key(e_aset *asp,e_name *key,int index)
 *                                 [0,0,.;.]
 **/
/**************************************************/ 
/* remove item from the aset by key.              */
/* asp : aset to modify                           */
/* key : key to match                             */
/* index : key accurence to remove                */
/*   n <= 0 : remove all occurence                */
/*   n >  0 : remove nth occurence only           */
/* The aset is not cloned!                        */
/* No link changed                                */
/* RETURN : 1 : OK                                */
/*          0 : error                             */
/**************************************************/ 
int e_aset_remove_item_by_key(e_aset *asp,e_name *key,int index)
{ int i;
  e_aset_item *aip,*prev,*next;

#ifdef E_DEBUG
  e_debug("e_aset_remove_item_by_key(asp=%x,key=%x,index=%d)\n",asp,key,index);
#endif


  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,
                            "e_aset_remove_item_by_key(asp,-,-)"));
  E_NULL_ERROR(e_check_type(key,E_TYPE_NAME,
                            "e_aset_remove_item_by_key(-,key,-)"));

  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* search item to remove */
  while(aip)
  { if(E_ASET_ITEM_NAME(aip) == key) 
    { i++;
      if((index <= 0) || (index == i))
      {           /* remove the item */
        prev = E_ASET_ITEM_PREV(aip);
        next = E_ASET_ITEM_NEXT(aip);

        if(prev) E_ASET_ITEM_NEXT(prev) = next;
        else     E_ASET_FIRST(asp) = next; 

        if(next) E_ASET_ITEM_PREV(next) = prev;
        else     E_ASET_LAST(asp) = prev; 

        E_ASET_CARDINAL(asp)--;
        aip = next;

        if(index == i) goto end; /* remove only one item */
      }
      else aip = E_ASET_ITEM_NEXT(aip);
    }
    else aip = E_ASET_ITEM_NEXT(aip);
  }

end:
  return(1);

error:
  return(0);  
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_diff_by_key(e_aset *asp1,e_aset *asp2,int max)
 *                            [-1,0,.;+1]
 **/
/**************************************************************************/
/* Return the difference by key of two aset.                              */
/* The difference is the set of all key from aset 1 not present in aset 2.*/
/* { a b c d } - { b f} = {a c d }                                        */
/* The key retain their value from aset 1                                 */
/* asp1 is cloned if it is used more than once.                           */
/* asp1 : aset to remove from.                                            */
/* asp2 : aset of object to remove.                                       */
/* max : maximum number of times a given object is removed (-1=>no limit) */
/* RETURN : aset1-aset2                                                   */
/* LINK : asp1 : -1                                                       */
/*        RETURN : +1                                                     */
/**************************************************************************/
e_aset *e_aset_diff_by_key(e_aset *asp1,e_aset *asp2,int max)
{ e_aset *new;
  e_aset_item *aip;
  e_name *key,*key2;
  int n;

#ifdef E_DEBUG
  e_debug("e_aset_diff_by_key(asp1=%x,asp2=%x,max=%d)\n",asp1,asp2,max);
#endif

  E_NULL_ERROR(e_check_type(asp1,E_TYPE_ASET,"e_aset_diff_by_key(asp,asp)"));
  E_NULL_ERROR(e_check_type(asp2,E_TYPE_ASET,"e_aset_diff_by_key(asp,asp)"));

  if (E_OBJ_REF(asp1) > 1)
  { new = e_aset_clone(asp1,E_CLONE_RECURSIVE);
    E_FREE(asp1);
    asp1 = new;
  }
  
  aip = E_ASET_FIRST(asp2);

  while(aip)
  { if((key2 = E_ASET_ITEM_NAME(aip)))
    { n = 0;
      e_aset_next(asp1,E_NEXT_RESET,0);
      while((e_aset_next(asp1,E_NEXT_GET_CUR_KEY,(void **)&key)>0)&&
	    ((max < 0) || (n < max)))
      { if(key == key2)
        { e_aset_remove_cur_item(asp1);
          n++;
        }
        e_aset_next(asp1,E_NEXT_NEXT,0);
      }
    }
    aip = aip->next;
  }
  return(asp1);

error:
  E_FREE(asp1);
  return(0);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_diff_by_value(e_aset *asp1,e_aset *asp2,int max)
 *                              [-1,0,.;+1]
 **/
/**************************************************************************/
/* Return the difference by key of two aset.                              */
/* The difference if the set of all the object from aset 1 not in aset 2. */
/* { a b c d } - { b f} = {a c d }                                        */
/* The entries retain the key from aset 1                                 */
/* asp1 is cloned if it is used more than once.                           */
/* RETURN : aset1-aset2                                                   */
/* asp1 : aset to remove from.                                            */
/* asp2 : aset of object to remove.                                       */
/* max : maximum number of times a given object is removed (-1=>no limit) */
/* LINK : asp1 : -1                                                       */
/*        RETURN : +1                                                     */
/**************************************************************************/
e_aset *e_aset_diff_by_value(e_aset *asp1,e_aset *asp2,int max)
{ e_aset *new;
  e_aset_item *aip;
  void *obj,*obj2;
  int n;

#ifdef E_DEBUG
  e_debug("e_aset_diff_by_value(asp1=%x,asp2=%x,max=%d)\n",asp1,asp2,max);
#endif

  E_NULL_ERROR(e_check_type(asp1,E_TYPE_ASET,"e_aset_diff_by_value(asp,asp)"));
  E_NULL_ERROR(e_check_type(asp2,E_TYPE_ASET,"e_aset_diff_by_value(asp,asp)"));

  if (E_OBJ_REF(asp1) > 1)
  { new = e_aset_clone(asp1,E_CLONE_RECURSIVE);
    E_FREE(asp1);
    asp1 = new;
  }
  
  aip = E_ASET_FIRST(asp2);

  while(aip)
  { if((obj2 = E_ASET_ITEM_VAL(aip)))
    { n = 0;
      e_aset_next(asp1,E_NEXT_RESET,0);
      while((e_aset_next(asp1,E_NEXT_GET_CUR,(void **)&obj)>0)&&
	    ((max < 0) || (n < max)))
      { if(obj && E_EQUALS(obj2,obj))
        { e_aset_remove_cur_item(asp1);
          n++;
        }
        e_aset_next(asp1,E_NEXT_NEXT,0);
      }
    }
    aip = aip->next;
  }
  return(asp1);

error:
  E_FREE(asp1);
  return(0);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_common_by_key(e_aset *asp1,e_aset *asp2)
 *                              [-1,0;+1]
 **/
/**************************************************************************/
/* Return the intersection between the two asets by key : the list of     */
/* items that are common to both asets. Multiple entries in aset1 are     */
/* retained while multiple entries in aset2 make no difference.           */
/* The key retain their value from aset 1                                 */
/* asp1 : main aset.                                                      */
/* asp2 : second aset.                                                    */
/* RETURN : interection of aset1 and aset2                                */
/* LINK : asp1 : -1                                                       */
/*        RETURN : +1                                                     */
/**************************************************************************/
e_aset *e_aset_common_by_key(e_aset *asp1,e_aset *asp2)
{ e_aset *new;
  e_aset_item *aip;
  e_name *key1,*key2;
  int n;

#ifdef E_DEBUG
  e_debug("e_aset_common_by_key(asp1=%x,asp2=%x)\n",asp1,asp2);
#endif

  E_NULL_ERROR(e_check_type(asp1,E_TYPE_ASET,"e_aset_common_by_key(asp,asp)"));
  E_NULL_ERROR(e_check_type(asp2,E_TYPE_ASET,"e_aset_common_by_key(asp,asp)"));

  new = e_aset_new(0);
  
  aip = E_ASET_FIRST(asp1);

  while(aip)
  { if((key1 = E_ASET_ITEM_NAME(aip)))
    { n = 0;
      e_aset_next(asp2,E_NEXT_RESET,0);
      while((e_aset_next(asp2,E_NEXT_GET_CUR_KEY,(void **)&key2)>0) && !n)
      { if(key1 == key2)
        {  E_NULL_ERROR(new = 
               e_aset_add_item(new,E_USE(key1),E_USE(E_ASET_ITEM_VAL(aip)),0));
          n++;
        }
        e_aset_next(asp2,E_NEXT_NEXT,0);
      }
    }
    aip = aip->next;
  }

  E_FREE(asp1);

  return(new);

error:
  E_FREE(asp1);
  return(0);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_common_by_value(e_aset *asp1,e_aset *asp2)
 *                                [-1,0;+1]
 **/
/**************************************************************************/
/* Return the intersection between the two asets by value : the list of   */
/* items that are common to both asets. Multiple entries in aset1 are     */
/* retained while multiple entries in aset2 make no difference.           */
/* The key retain their value from aset 1                                 */
/* asp1 : main aset.                                                      */
/* asp2 : second aset.                                                    */
/* RETURN : interection of aset1 and aset2                                */
/* LINK : asp1 : -1                                                       */
/*        RETURN : +1                                                     */
/**************************************************************************/
e_aset *e_aset_common_by_value(e_aset *asp1,e_aset *asp2)
{ e_aset *new;
  e_aset_item *aip;
  void *obj1,*obj2;
  int n;

#ifdef E_DEBUG
  e_debug("e_aset_common_by_value(asp1=%x,asp2=%x)\n",asp1,asp2);
#endif

  E_NULL_ERROR(e_check_type(asp1,E_TYPE_ASET,
                            "e_aset_common_by_value(asp,asp)"));
  E_NULL_ERROR(e_check_type(asp2,E_TYPE_ASET,
                            "e_aset_common_by_value(asp,asp)"));

  new = e_aset_new(0);
  
  aip = E_ASET_FIRST(asp1);

  while(aip)
  { if((obj1 = E_ASET_ITEM_VAL(aip)))
    { n = 0;
      e_aset_next(asp2,E_NEXT_RESET,0);
      while((e_aset_next(asp2,E_NEXT_GET_CUR,(void **)&obj2)>0) && !n)
      { if(obj2 && E_EQUALS(obj2,obj1))
        { E_NULL_ERROR(new = 
             e_aset_add_item(new,E_USE(E_ASET_ITEM_NAME(aip)),E_USE(obj1),0));
          n++;
        }
        e_aset_next(asp2,E_NEXT_NEXT,0);
      }
    }
    aip = aip->next;
  }

  E_FREE(asp1);

  return(new);

error:
  E_FREE(asp1);
  return(0);
}

/***************************** TEST ASET **********************************/
/*****************************************************************************/
/**
 *  int e_aset_is_in(e_aset *asp,e_name *key,void *val)
 *                   [0,0,0;.]
 **/
/**********************************************************/
/* Check if an item key/value  belong to the aset.        */
/* asp : aset to check                                    */
/* key : key to search in aset. If NULL only test val.    */
/* val : value to match. If NULL only test key.           */
/* If both key and val are non null, they must both match */
/* RETURN : 0 : item not found                            */
/*          1 : item is a member.                         */
/*          -1 : key found but with the wrong value       */
/* LINK : no change.                                      */
/**********************************************************/
int e_aset_is_in(e_aset *asp,e_name *key,void *val)
{ e_aset_item *aip;
  int ret = 0;

#ifdef E_DEBUG
  e_debug("e_aset_is_in(asp=%x,key=%x,val=%x)\n",asp,key,val);
#endif

  if(!key && !val) return(0);

  if(key)
  { aip = E_ASET_FIRST(asp);
  
    while(aip)
    { if(E_ASET_ITEM_NAME(aip) == key) 
      { if (!val || E_EQUALS(E_ASET_ITEM_VAL(aip),val)) return(1);
        ret = -1;
      }
      aip = E_ASET_ITEM_NEXT(aip);
    }

    return(ret);
  }
  else
  { if(e_aset_get_value_index(asp,E_USE(val),0) < 0) return(0);
    else return(1); 
  }
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_split(e_string *sp,char *sep, char *psep)
 *                      [-1,.,.;+1]
 **/
/************************************************************************/
/* Split a string into a set of words. sep are the list of space        */
/* separators (as a string), while psep are the punctuation separators. */
/* By default sep = " \t" and psep = "".                                */
/*                                                                      */
/* Examples: sep= " ", psep = ",;"                                      */
/* "a b c"      => { "a" "b" "c" }                                      */
/* "a,b,c"      => { "a" "b" "c" }                                      */
/* "a,,c"       => { "a" "" "c" }                                       */
/* "a c"        => { "a" "c" }                                          */
/* "a , b , c"  => { "a" "b" "c" }                                      */
/* sp : string to split                                                 */
/* sep : space separators. Multiple space only count as one             */
/* psep : punctuation separator. Each one marks then end of a word      */
/* Return : the set of strings.                                         */
/*          0 : error                                                   */
/* Link : sp :     -1                                                   */
/*        Return : +1                                                   */
/************************************************************************/
e_aset *e_aset_split(e_string *sp,char *sep, char *psep)
{ e_tokeniser *tp = 0;
  e_aset *asp = 0;
  int state,r;

#ifdef E_DEBUG
  e_debug("e_aset_split(sp=%x,sep=%s,psep=%s)\n",sp,sep,psep);
#endif

#define E_ASET_SPLIT_MODE_WORD 1
#define E_ASET_SPLIT_MODE_PUNC 2

  if(!sep) sep = " \t";
  if(!psep) psep = "";
  
  if(!(tp = e_tokeniser_new(0,e_charstream_new(0,
                                      E_CHARSTREAM_TYPE_STRING,sp,E_MODE_READ),
                            0," -~","",psep,sep)))
  { goto error;
  }

  E_NULL_ERROR(asp = e_aset_new(0)); 

  state = E_ASET_SPLIT_MODE_WORD;
  while((r = e_tokeniser_next(tp)) > 0)
  { 
    switch(tp->type)
    { case E_TOKENISER_WORD_TOKEN :
	                               /* Add word to set*/
           E_NULL_ERROR(asp = e_aset_append_item(asp,0,
                               e_string_new(0,E_TOKENISER_BUFF(tp),0)));  
           state = E_ASET_SPLIT_MODE_WORD; 
           break;

      case E_TOKENISER_OP_TOKEN :       /* read the = op */
           if(state == E_ASET_SPLIT_MODE_PUNC)
	   {                            /* Add empty string to set */
             E_NULL_ERROR(asp = e_aset_append_item(asp,0,
                               e_string_new(0,"",0))); 
           }

           state = E_ASET_SPLIT_MODE_PUNC;
           break;
    }
  }
  if(!r) goto error;

  return(asp);

error:
  if(tp)   E_FREE(tp);
  if(asp)  E_FREE(asp);
  return(0);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_add_property(e_aset *asp, e_name *key, void *val,int mode)
 *                             [-1,-1,-1,.;+1]
 **/
/**********************************************************************/ 
/* Add the property val (usualy a string but is can be any object)    */
/* to the property key.                                               */
/* asp is a list of properties, each identified by a key.             */
/* Each property list is an aset of objects.                          */
/* If the property is not in the list yet, it is added to asp. The    */
/* property list is the aset containing the single object obj. If val */
/* is null the property list is the empty aset.                       */
/* If the property key already exists its value is added to the       */
/* property list after checking that val is not already in the        */
/* property list.                                                     */
/* asp : a aset of aset : the aset of properties.                     */
/* key : a e_name : the property name (key in the aset)               */
/* mode : E_ASET_PROPERTY_ADD_SET : val is a set of properties to add */
/*        E_ASET_PROPERTY_MULTI : do not check if val is already in   */
/*               the property list (allow multiplicity values)        */
/* Return : the new property list.                                    */
/*          0 :error                                                  */
/* Link : asp  -1                                                     */
/*        key : -1                                                    */
/*        val : -1                                                    */
/*        Return : +1                                                 */
/**********************************************************************/ 
e_aset *e_aset_add_property(e_aset *asp, e_name *key, void *val,int mode)
{ e_aset *p,*aspv=0;
  e_aset_item *aip;

#ifdef E_DEBUG
  e_debug("e_aset_add_property(asp=%x,key=%x,val=%x,mode=%x)\n",
          asp,key,val,mode);
#endif

  if(!asp) 
  { e_error_null_pointer("e_aset_add_property(asp,-,-,-)");  
    goto error;
  }
  if(!key) 
  { e_error_null_pointer("e_aset_add_property(-,key,-,-)"); 
    goto error;
  }

             /* Make sure the aset is used only once     */
             /* So that we can modify it directly below! */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }


  if(!(aip = e_aset_get_item_by_key(asp,E_USE(key),1)))
  {                      /* Key not found : add  key = { val }  */
    return(e_aset_append_item(asp,key,
                              e_aset_append_item(e_aset_new(0),0,val)));
  }

  if(E_OBJ_READ_ONLY(aip))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_USER,
            "%s : is write protected",E_OBJ_NAME(key));
    return(0);
  }

              /* Is val already in the set ? */
  aspv = E_ASET_ITEM_VAL(aip);
  E_NULL_ERROR(e_check_type(aspv,E_TYPE_ASET,
      "e_aset_add_property : property item must be a set."));
  

  if(mode && E_ASET_PROPERTY_ADD_SET)
  { E_NULL_ERROR(e_check_type(val,E_TYPE_ASET,
      "e_aset_add_property(-,val,ADD_SET) : val must be a set."));
 
    if(mode & E_ASET_PROPERTY_MULTI)
    {                       /* Just append the list */
      E_ASET_ITEM_VAL(aip) = e_aset_append(aspv,val);
      return(asp);
    }
                            /* Add each element one by one. */
    E_ASET_ITEM_VAL(aip) = e_aset_add_aset(E_ASET_ITEM_VAL(aip),val,
                                             E_ASET_ADD_MODE_UNIQUE_VAL);
    return(asp);
  }


  if((mode & E_ASET_PROPERTY_MULTI) || 
     (e_aset_get_value_index(aspv,E_USE(val),0) < 0))
  {                      /* Property not yet in the set : add it */
    E_ASET_ITEM_VAL(aip) = e_aset_append_item(aspv,0,val);  
  }  
  return(asp);

error:
  if(aspv) E_FREE(aspv);  
  if(key) E_FREE(key);  
  if(val) E_FREE(val);  
  return(0);
}

/*****************************************************************************/
/**
 *  e_aset *e_aset_get_properties(e_aset *asp, void *val)
 *                               [-1,-1;+1]
 **/
/**********************************************************************/ 
/* Create a subset of the property list for all the properties which  */
/* have the value (val) set.                                          */
/* The aset returned is a subset of the property list:                */
/* asp : a aset of aset : the aset of properties.                     */
/* key : a e_name : the property name (key in the aset)               */
/* Return : the aset of set properties.                               */
/*          0 :error                                                  */
/* Link : asp  -1                                                     */
/*        val : -1                                                    */
/*        Return : +1                                                 */
/**********************************************************************/ 
e_aset *e_aset_get_properties(e_aset *pl, void *val)
{ e_aset *asp=0,*plist;
  e_aset_item *aip;
  e_name *pname;

  if(!pl || !val || !(asp = e_aset_new(0))) goto error;

  aip = E_ASET_FIRST(pl);
  
  while(aip)
  { plist = E_ASET_ITEM_VAL(aip);
    pname = E_ASET_ITEM_NAME(aip);

    if(pname && plist && E_IS_ASET(plist) && 
       (e_aset_get_value_index(plist,E_USE(val),0) >= 0))
    { E_NULL_ERROR(asp = e_aset_append_item(asp,E_USE(pname),E_USE(plist)));
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  E_SAFE_FREE(pl);
  E_SAFE_FREE(val);
  return(asp);

error:
  E_SAFE_FREE(asp);
  E_SAFE_FREE(pl);
  E_SAFE_FREE(val);
  return(0);
}

/*****************************************************************************/
/**
 *   e_aset *make_aset(void *vals, ...)
 *                             [-1,-1, ... ,-1;+1]
 **/
/********************************************/
/* Create an arbitrary size aset (anomyous) */
/* make_aset(ob1, ob2, ...,0)               */
/* All obi must be easy objects             */
/* Return : new aset.                       */
/*          0 : error                       */
/* Link : -1 : obi                          */
/*        +1 : RETURN                       */
/********************************************/
e_aset *make_aset(void *ob1, ...)
{ va_list args;
  void *p;
  e_aset *asp;

  E_NULL_ERROR(asp = e_aset_new(0));

  va_start(args, ob1);
  p = ob1;
  do
  {if(!(asp = e_aset_append_item(asp,0,p)))
    { va_end(args);
      goto error;
    }
  }
  while((p = va_arg(args,void *)));

  va_end(args);

  return(asp);

error:
  E_SAFE_FREE(asp);
  return(0);
}

/*****************************************************************************/
/**
 *   e_aset *make_named_aset(void *n1, ...)
 *                             [-1,-1, ... ,-1;+1]
 **/
/*****************************************************/
/* Create an arbitrary size aset (anomyous)          */
/* make_named_aset(n1,ob1, n2, ob2, ...,0)           */
/* All ni must ne named and obi must be easy objects */
/* The ni AND obi CAN'T be NULL!                     */
/* Return : new aset.                                */
/*          0 : error                                */
/* Link : -1 : ni,obi                                */
/*        +1 : RETURN                                */
/*****************************************************/
e_aset *make_named_aset(void *n1, ...)
{ va_list args;
  void *p,*ob;
  e_aset *asp;

  if(!(asp = e_aset_new(0))) return(0);
  
  va_start(args, n1);

  p = n1;
  do
  { if(!p || !E_IS_NAME(p) || !(ob = va_arg(args,void *))) goto error;
    E_NULL_ERROR(asp = e_aset_append_item(asp,p,ob));
  }
  while((p = va_arg(args,void *)));

  va_end(args);

  return(asp);

error:
  va_end(args);
  E_SAFE_FREE(asp);
  return(0);
}
