#include <easy.h>
#include <stdlib.h>

e_obj_descriptor e_obj_descriptor_name = {
  "name",             /* Name      */
  E_OBJ_MODE_NO_CLONE,/* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_name_free,        /* free      */
  e_name_to_string,   /* to_string */
  e_name_clone,       /* clone     */
  e_name_equals,      /* equals    */
};

                    /* Hash table of names */
                    /* Names with the same hash values are linked together. */
e_name *e_name_table[E_NAME_TABLE_SIZE];

/***************************/
/* Initialisation function */
/***************************/
int e_name_init(void *p)
{ int i;

           /* Set each name entry to NULL */
  for(i = 0; i < E_NAME_TABLE_SIZE; i++)
  { e_name_table[i] = 0;
  }

  return(1);
}


/************************************************************/
/* NAME :                                                   */
/* Global entry for a name.                                 */
/* Each string has a unique name entry.                     */
/* Names are never cloned.                                  */
/* All the names are stored in a global hash table          */
/************************************************************/
/***************************************************************************/
/* Create or get a new e_name                                              */
/* Name can't be static. If a name already exist, the pointer to that      */
/* existing name is returned.                                              */
/* The name is added to the hash table (with a link) and is returned       */
/* with a link too.                                                        */
/* Return e_name                                                           */
/*        0 : error                                                        */
/* Link : Return +1                                                        */
/***************************************************************************/
e_name *e_name_new(char *name)
{ char *sp = 0;
  e_name *np = 0,*tnp;
  int hash;

                 /* Static object or allocate it ? */
  E_NULL_ERROR(name);
  hash = e_name_hash(name,E_NAME_TABLE_SIZE);
         
                 /* Does the name exists already? */
                 /* If so, return it with a link  */
  if((tnp = e_name_table[hash]))
  { do 
    { if(!strcmp(E_NAME_NAME(tnp),name)) { return(E_USE(tnp)); }
    } while((tnp = tnp->next));
  }

                 /* New name => allocate it       */
  if (!(np = e_obj_descriptor_name._alloc(sizeof(e_name)))) 
      return((e_name *)e_error_no_memory("e_name_new"));

  np->h.obj_desc_p = &e_obj_descriptor_name;
  np->h.refs = 2;
  np->h.mode = 0;
  np->val = 0;
  np->next = 0;
  np->hash = hash;

                 /* Allocate Name buffer */
  np->len = strlen(name);
  if (!(sp = e_obj_descriptor_name._alloc(np->len+1)))
  { e_error_no_memory("e_string_new");
    goto error;
  }
  np->name = sp;
  strcpy(np->name,name);


                 /* Add the name to the hash table. Name with the same    */
                 /* hash value are linked together.                       */
                 /* The new name is inserted at the beginning of the list */
  tnp = e_name_table[hash];
  e_name_table[hash] = np;
  np->next = tnp;

  return(np);

error:
  if(sp) e_obj_descriptor_name._free(sp);
  if(np->h.refs > 0) e_obj_descriptor_name._free(np);
  return(0); 
}

/***************************************************************************/
/* Free the name                                                           */
/* Names can't be freed as they keep at least a link in the hash table     */
/* So the ref counter is decremented.                                      */
/* If the Ref counter is 1, an error occurs.                               */
/* REF : np : -1                                                           */
/***************************************************************************/
void e_name_free(void *np)
{

  if(np)
  { if(E_OBJ_REF(np) <= 1)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a name (%x : %s) ",np,E_NAME_NAME(np));
      goto error;
    }

    E_OBJ(np)->refs--; 
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/*************************************************************/
/* Convert the name to a string                              */
/* The name is converted to a string                         */
/* np : a name, assumed to be of the correct type            */
/* tab: left identation (ignored for COMPACT mode)           */
/* mode:  0, E_TO_STR_COMPACT                                */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : np : 0                                             */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_name_to_string(void *np,int tab,e_flag mode)
{ e_string *stp = 0;

  E_NULL_ERROR(stp = e_string_new(0,E_NAME_NAME(np),E_NAME_LEN(np)));

  if(!(mode & E_TO_STR_NO_TAB))
  { /* add tab to the string */
    stp = e_string_insert_chars(stp,0,tab,' ');
  }

  return (stp);

error:
  if (stp) E_FREE(stp);
  return(0);
}

/*********************************************************************/
/* Clone a name                                                      */
/* A name can never be cloned => return the name with one extra link */
/* Return : np                                                       */
/*          0 : Error                                                */
/* LINK : RETURN : +1                                                */
/*********************************************************************/
void *e_name_clone(void *np,e_flag mode)
{ 
  if(np)
  { E_OBJ(np)->refs++;
  } 
  return(np);
}

/*************************************************************/
/* Test if the 2 names are equal:                            */
/* The two names are equal if they are identical (p1 == p2). */
/* Return : 1 : p1 and p2 are equal                          */
/*          0 : p1 and p2 are different                      */
/* Link : p1 0                                               */
/*        p2 0                                               */
/*************************************************************/
int e_name_equals(void *p1,void *p2)
{ 

  if (p1 == p2)   return(1);

  return(0);
}

/*********************************************/
/* Does the same as e_name_new but return    */
/* the name without a link                   */
/* char *s : name to find                    */
/* Return : name                             */
/*          0 : error                        */ 
/* Link : Return 0                           */
/*********************************************/
e_name *e_name_find_name(char *s)
{ e_name *np;

  if((np = e_name_new(s)))
  { E_OBJ_REF(np)--;          /* Free the link returned by e_name_new */
  }

  return(np);
}

/*********************************************/
/* Set the global value of the name to val.  */
/* np  : name to set                         */
/* val : new value for the name              */
/* if val is NULL, the name value is errased */
/* Return : 1 : OK                           */
/*          0 : error                        */ 
/* Link : np  -1                             */
/*        val -1                             */
/*********************************************/
int e_name_set_val(e_name *np,void *val)
{   
  E_NULL_ERROR(np);

  if(E_OBJ_READ_ONLY( np))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_USER,
            "Can't modify name \"%s\"!",E_NAME_NAME(np)
           );
    goto error;
  }

  if(E_NAME_VAL(np)) E_FREE(E_NAME_VAL(np));

  E_NAME_VAL(np) = val;
  E_FREE(np);

  return(1);  

error:
  if(val) E_FREE(val);
  if(np) E_FREE(np);
  return(0);
}

/*********************************************/
/* Get the global value of the name to val.  */
/* np  : name                                */
/* Return : Value                            */
/*          0 : error, or no value           */ 
/* Link : np  -1                             */
/*        Return +1                          */
/*********************************************/
void *e_name_get_val(e_name *np)
{ void *val = 0;
  
  E_NULL_ERROR(np);

  if(E_OBJ_NO_READ(np))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_USER,
            "Tryng to read a Read proected Name \"%s\"",E_NAME_NAME(np)
           );
    goto error;
  }

  if((val = E_NAME_VAL(np))) E_USE(val);
  E_FREE(np);

  return(val);  

error:
  if(val) E_FREE(val);
  if(np) E_FREE(np);
  return(0);
}

/*****************************************/
/* Compute the hash value of a string    */
/* s : string. (must be non NULL))       */
/* hashsize : size of the Hash Table     */
/* hashsize must be less than 2^24       */
/* return : Hash value (0 -> hashsize-1) */
/*                                       */
/*****************************************/
int e_name_hash(char *s, int hashsize)
{ register int h;
  register int c;

  h = 0;

  while ((c = *s++))
  { h = (h << 3) -h + c; /* h => h*7 + c */
    /* Feed high bit back in */
    h = (h+(h >> 24)) && 0xffffff;   
  }

  return ((int) h % hashsize);
}
