#include <easy.h> 

e_obj_descriptor e_obj_descriptor_tokeniser = {
  "parser",             /* Name      */
  E_OBJ_MODE_NO_CLONE,  /* def_mode  */
  malloc,               /* _alloc    */
  free,                 /* _free     */
  realloc,              /* _realloc  */
  e_tokeniser_free,        /* free      */
  e_tokeniser_to_string,   /* to_string */
  e_tokeniser_clone,       /* clone     */
  e_tokeniser_equals,      /* equals    */
};


/**************************************************************/
/* TOKENISER : tokenise a charstream and convert it to tokens.*/
/***********************************************************/
/* Typical usage                                           */
/*                                                         */
/***********************************************************/

/******************* TOKENISER STANDARD METHODS  ************************/

/***************************************************************************/
/* Create or initialise a new e_tokeniser                                  */
/* pp : e_tokeniser to initialise (static e_tokeniser)                     */
/*      if NULL allocate it                                                */
/*      pp is assumed to be of the correct type                            */
/* csp : the charstream to read from.                                      */
/* words  : set of valid word characters                                   */
/* op     : set of valid operator characters                               */
/* punct  : set of valid punctuation characters                            */
/* spaces : set of valid spaces characters                                 */
/* When words, op, punct or spaces are null the default values are used.   */
/* Return e_tokeniser                                                      */
/*        0 : error                                                        */
/* Link : csp -1                                                           */
/*        Return +1                                                        */
/***************************************************************************/
e_tokeniser *e_tokeniser_new(e_tokeniser *tp,e_charstream *csp,int mode,
                      char *word,char *op,char *punct, char *space)
{ int j;

  if(!csp) return(0);

                 /* Static object or allocate it ? */
  if (!tp)
  { if (!(tp = e_obj_descriptor_tokeniser._alloc(sizeof(e_tokeniser)))) 
      return((e_tokeniser *)e_error_no_memory("e_tokeniser_new"));
    tp->h.refs = 1;
  }
  else tp->h.refs = -1;
  tp->h.mode = 0;

  tp->h.obj_desc_p = &e_obj_descriptor_tokeniser;

                /* is mode non zero ? */
  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,
              "e_tokeniser_new(-,csp,-,-,-,-,-)"));

  E_TOKENISER_CHARSTREAM(tp) = csp;

                 /* Initialise types to BIN by default */
  for(j=0; j < 256; j++) 
  { E_TOKENISER_TYPES(tp)[j] = (char) E_TOKENISER_BIN_CHAR;
  }

  if (!word)   word   = E_TOKENISER_DEF_WORD;
  if (!op)     op     = E_TOKENISER_DEF_OP;
  if (!punct)  punct  = E_TOKENISER_DEF_PUNCT;
  if (!space)  space  = E_TOKENISER_DEF_SPACES;
  E_NULL_ERROR(e_tokeniser_set_char_types(E_TOKENISER_TYPES(tp),
                                        word,E_TOKENISER_WORD_CHAR));
  E_NULL_ERROR(e_tokeniser_set_char_types(E_TOKENISER_TYPES(tp),
                                        op,E_TOKENISER_OP_CHAR));
  E_NULL_ERROR(e_tokeniser_set_char_types(E_TOKENISER_TYPES(tp),
                                        punct,E_TOKENISER_PUNCT_CHAR));
  E_NULL_ERROR(e_tokeniser_set_char_types(E_TOKENISER_TYPES(tp),
                                        space,E_TOKENISER_SPACE_CHAR));

  tp->sp = 0;
  tp->Buff[0] = 0;
  tp->mode = mode;
  tp->type = 0;
  tp->newline = 1;
  tp->line_no = 1;

  return(tp);

error:
  if(csp) E_FREE(csp);
  if(tp && (tp->h.refs > 0)) e_obj_descriptor_tokeniser._free(tp);
  return(0); 
}

/***************************************************************************/
/* Free and close the tokeniser                                            */
/* If the refs counter is -1 : static tokeniser : close the tokeniser and  */
/*                             free the charstream.                        */
/* If the refs counter is 1 : free the allocated memory for the tokeniser  */
/* Otherwise decrement the counter by 1.                                   */
/* REF : stp : -1                                                          */
/***************************************************************************/
void e_tokeniser_free(void *tp)
{

  if(tp)
  { if(E_OBJ_REF(tp) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free tokeniser (%x) ",tp);
      goto error;
    }

    if (E_OBJ_REF(tp) < 0)      /* static tokeniser, free the charstream */
    { if (E_TOKENISER_CHARSTREAM(tp)) 
      { E_FREE(E_TOKENISER_CHARSTREAM(tp));
        E_TOKENISER_CHARSTREAM(tp) = 0;
      }
      return;
    }

                                /* decrement the refs counter   */
    if (E_OBJ(tp)->refs-- > 1) return;

                                /* Last reference, 
                                   free the file filename  */
    if (E_TOKENISER_CHARSTREAM(tp)) E_FREE(E_TOKENISER_CHARSTREAM(tp));

                                /* free the e_tokeniser */
    e_obj_descriptor_file._free(tp);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/*************************************************************/
/* Convert the tokeniser to a string                         */
/* The tokeniser is converted to the string                  */
/* Tokeniser(CHARSTREAM,MODE,WORDS,OP,PUNCT,SPACE )          */
/* tp : a tokeniser, assumed to be of the correct type       */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : tp : 0                                             */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_tokeniser_to_string(void *tp,int tab,e_flag mode)
{ e_string *stp = 0;
  char buff[E_LINE_SIZE+2];

  E_NULL_ERROR(stp = e_string_new(0,"Tokeniser(",E_LINE_SIZE));

  E_NULL_ERROR(stp = e_string_append_string(stp,
               E_TO_STRING(E_TOKENISER_CHARSTREAM(tp),0,
                           E_TO_STR_COMPACT|E_TO_STR_QUOTE)));

  buff[0] = ',';
  e_tokeniser_char_types_to_buffer(E_TOKENISER_TYPES(tp),
                                   buff+1,E_TOKENISER_WORD_CHAR);
  E_NULL_ERROR(stp = e_string_append_buffer(stp,buff));

  buff[0] = ',';
  e_tokeniser_char_types_to_buffer(E_TOKENISER_TYPES(tp),
                                   buff+1,E_TOKENISER_OP_CHAR);
  E_NULL_ERROR(stp = e_string_append_buffer(stp,buff));

  buff[0] = ',';
  e_tokeniser_char_types_to_buffer(E_TOKENISER_TYPES(tp),
                                   buff+1,E_TOKENISER_PUNCT_CHAR);
  E_NULL_ERROR(stp = e_string_append_buffer(stp,buff));

  buff[0] = ',';
  e_tokeniser_char_types_to_buffer(E_TOKENISER_TYPES(tp),
                                   buff+1,E_TOKENISER_SPACE_CHAR);
  strcat(buff,")");
  E_NULL_ERROR(stp = e_string_append_buffer(stp,buff));

  if(!(mode & E_TO_STR_NO_TAB))
  { /* add tab to the string */
    stp = e_string_insert_chars(stp,0,tab,' ');
  }

  return (stp);

error:
  if (stp) E_FREE(stp);
  return(0);
}

/*****************************************************************/
/* Clone a tokeniser                                             */
/* mode : cloning mode                                           */
/* If the tokeniser can't be cloned (default), return tp         */
/* and increase its refs counter                                 */
/* If cloned, the tokeniser is just reopened with the same       */      
/* charstream reopenned too. This would rewind the charstream.   */
/* tp : a tokeniser, assumed to be of the correct type           */
/* Return : cloned e_tokeniser                                   */
/*          0 : Error                                            */
/* LINK : stp : 0                                                */
/*        RETURN : +1                                            */
/*****************************************************************/
void *e_tokeniser_clone(void *tp,e_flag mode)
{ e_tokeniser *newtp = 0;
  e_charstream *newcs = 0;
  e_flag obj_mode;
  int i;

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(tp)) return(E_USE(tp));

                 /* save the charstream mode */
  obj_mode = E_OBJ_MODE(E_TOKENISER_CHARSTREAM(tp));

                 /* make the charstream clonable */
  E_OBJ_MODE(E_TOKENISER_CHARSTREAM(tp)) &= ~E_OBJ_MODE_NO_CLONE;

                 /* clone the charstream an create a new tokeniser */
  if(!(newcs = E_CLONE(E_TOKENISER_CHARSTREAM(tp),mode)))
  { E_OBJ_MODE(E_TOKENISER_CHARSTREAM(tp)) = obj_mode;
    goto error;
  }

                 /* restore the charstream mode */
  E_OBJ_MODE(E_TOKENISER_CHARSTREAM(tp)) = obj_mode;

                 /* create a new tokeniser from the cloned object */
                 /* The char type are copied afterward            */
  if(!(newtp = e_tokeniser_new(0,newcs,E_TOKENISER_MODE(tp),0,0,0,0)))
  { newcs = 0;
    goto error;
  }
                 /* copy all the parameters */
  for(i=0; i < 256; i++)
  { E_TOKENISER_TYPES(newtp)[i] = E_TOKENISER_TYPES(tp)[i];
  }

  return(newtp);

error:
  if(newcs) E_FREE(newcs); 
  if(newtp) E_FREE(newtp); 
  return(0);
}

/******************************************************************/
/* Test if the 2 tokenisers are equal:                            */
/* The two tokenisers are equal if they are identical (p1 == p2), */
/* or if they have the same charstream, the same mode and the     */
/* list of character types.                                       */
/* The status are not compared!                                   */
/* p1 is assumed to be of type e_tokeniser                        */
/* Return : 1 : p1 and p2 are equal                               */
/*          0 : p1 and p2 are different                           */
/* Link : p1 0                                                    */
/*        p2 0                                                    */
/******************************************************************/
int e_tokeniser_equals(void *p1,void *p2)
{ int i; 

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

            /* different types => differ */
  if(E_OBJ_DESCR(p1) != E_OBJ_DESCR(p2)) return(0);

            /* different modes => differ */
  if(E_TOKENISER_MODE(p1) != E_TOKENISER_MODE(p1)) return(0);

  for(i=0; i < 256; i++)
  { if(E_TOKENISER_TYPES(p1)[i] != E_TOKENISER_TYPES(p2)[i]) return(0);
  }
            /* then compare the charstream */
  return(E_EQUALS(E_TOKENISER_CHARSTREAM(p1),E_TOKENISER_CHARSTREAM(p2)));

}


/******************* TOKEN FUNCTIONS ************************/

/*************************************/
/* Return : the type of token read   */
/*          -1 : No More Token       */
/*          0 : error                */
/*************************************/
int e_tokeniser_next(e_tokeniser *tp)
{ int start=0,i,j,c,d,loop;
  char ch[3];
  e_charstream *csp;

  csp = E_TOKENISER_CHARSTREAM(tp);

  c = E_CHARSTREAM_GET(csp);

  do
  { if(c == '\n') 
    { tp->line_no++; 

                       /* Return end of line token ? */
      if(E_TOKENISER_MODE(tp) & E_TOKENISER_MODE_RETRUN_EOL)
      { E_TOKENISER_BUFF(tp)[0] = 0;
        tp->type = E_TOKENISER_EOL_TOKEN;
        return(E_TOKENISER_EOL_TOKEN);
      }
                      /* if not get the next char */
      c = E_CHARSTREAM_GET(csp);
      tp->newline = 1;
    }

                       /* skip lines staring with # ? */
    while((c == '#') && tp->newline && 
          (tp->mode & E_TOKENISER_MODE_HASH_COMMENT))
    { while(((c = E_CHARSTREAM_GET(csp)) != '\n')&&(c != EOF));
      tp->line_no++; 
      tp->newline = 1;
      c = E_CHARSTREAM_GET(csp);
    }

    if(c == EOF) goto EndOfFile; 
 
                       /* skip spaces */
    while(E_TOKENISER_CHAR_IS_SPACE(tp,c)) c = E_CHARSTREAM_GET(csp);

    if(c == '\\') 
    {                  /* Continuation character : read up to end of line */
      if (tp->mode & E_TOKENISER_MODE_SLASH_IS_CONT) 
      {                /* read all characters up to the end of line */
        while (((c = E_CHARSTREAM_GET(csp))!= EOF) && (c !='\n')); 
        tp->line_no++;
      }

                      /* // Comment : read up to end of line */
      else if(tp->mode & E_TOKENISER_MODE_SLASH_SLASH_COMMENT)
      { 
        if((d = E_CHARSTREAM_GET(csp)) == '\\')
        {             /* read all characters up to the end of line */
          while (((c = E_CHARSTREAM_GET(csp))!= EOF) && (c !='\n'));
          tp->newline = 1;
          tp->line_no++;
        }
        else if (d != EOF) E_CHARSTREAM_UNGET(csp,d);
      }
    }

    if(c == EOF) goto EndOfFile; 
  }
  while(c == '\n');     /* empty line : start again */


              /******* PARSE A PUNCTUATION CHAR  **********/
  tp->newline = 0;
  i = 0;
  if(E_TOKENISER_CHAR_IS_PUNCT(tp,c))     /* Punctuation character */
  { tp->Buff[0] = c;
    tp->Buff[1] = 0;
    tp->type = E_TOKENISER_PUNCT_TOKEN;
    return(E_TOKENISER_PUNCT_TOKEN);
  }

              /******* PARSE A QUOTED STRING  **********/
  if((c == '"') && (tp->mode & E_TOKENISER_MODE_READ_STRING)) 
  { i = 0;  
    start = tp->line_no;
    while (((c = E_CHARSTREAM_GET(csp))!= EOF) && (c !='"'))
    { if(c == '\\')
      { if((c = E_CHARSTREAM_GET(csp)) == EOF) goto error_EOF; 
        switch(c)
        { case 'a' : c = '\a'; break;
          case 'b' : c = '\b'; break;
          case 'f' : c = '\f'; break;
          case 'n' : c = '\n'; break;
          case 'r' : c = '\r'; break;
          case 't' : c = '\t'; break;
          case 'v' : c = '\v'; break;
          case '0' : c = 0;    break;

          case '\n' : /*   \ at EOL removes the \n */
              if((c = E_CHARSTREAM_GET(csp)) == EOF) goto error_EOF; 
              break;                          
        }                          
      }
      else if(c == '\n') tp->line_no++;
      tp->Buff[i++] = c;
    } 

    tp->type = E_TOKENISER_STRING_TOKEN;
    tp->Buff[i] = 0;
    return(E_TOKENISER_STRING_TOKEN);
  }

              /******* PARSE A QUOTED CHAR  **********/
  if((c == '\'') && (tp->mode & E_TOKENISER_MODE_READ_CHAR)) 
  { i = 0;  
    if((c = E_CHARSTREAM_GET(csp)) == EOF) goto error_Bad_Quot_Char;

    if(c == '\\')
    { if((c = E_CHARSTREAM_GET(csp)) == EOF) goto error_Bad_Quot_Char; 


      if(('0' <= c) && (c <= '7')) /* Octal constant */
      { d = 0;
        ch[0] = ch[1] = ch[2] = ' ';
        for(j = 0; j < 3; j++)
        { if((c = E_CHARSTREAM_GET(csp)) == EOF) goto error_Bad_Quot_Char; 
          ch[j] = c;
          if(('0' <= c) && (c <= '7')) d = d*8+c-'0';
          else 
          { e_error_syntax(
                "Line %d : invalid character constant : `\\x%c%c%c` \n",
                tp->line_no,ch[0],ch[1],ch[2]);           
                return(0);  
          }
        }
        c = d;
      }
      else
      { switch(c)
        { case 'a' : c = '\a'; break;
          case 'b' : c = '\b'; break;
          case 'f' : c = '\f'; break;
          case 'n' : c = '\n'; break;
          case 'r' : c = '\r'; break;
          case 't' : c = '\t'; break;
          case 'v' : c = '\v'; break;
          case '0' : c = 0;    break;
          case '\n' : /*  Error */
               goto error_Bad_Quot_Char;
                                  
          case 'x' : /* Hexadecinal constant */
            d = 0;
            ch[0] = ch[1] = ' ';
            for(j = 0; j < 2; j++)
            { if((c = E_CHARSTREAM_GET(csp)) == EOF) goto error_Bad_Quot_Char; 
              ch[j] = c;

              if(('0' <= c) && (c <= '9')) d = d*16 + c- '0';
              else if (('a' <= c) && (c <= 'f')) d = d*16+c+10-'a';  
              else 
              { e_error_syntax(
                  "Line %d : invalid character constant : `\\x%c%c` \n",
                  tp->line_no,ch[0],ch[1]);           
                return(0);  
              }
            }
            c = d; 
            break;          
        }
      }
    }

    if(c == EOF) goto error_EOF;
    tp->Buff[i++] = c;

    if((c = E_CHARSTREAM_GET(csp)) != '\'') goto error_Bad_Quot_Char;

    tp->type = E_TOKENISER_CHAR_TOKEN;

    tp->Buff[i] = 0;
    return(E_TOKENISER_CHAR_TOKEN);
  } 


              /******* PARSE A WORD/OPERATOR  **********/
  i = 0;
  if(E_TOKENISER_CHAR_IS_OP(tp,c)) tp->type = E_TOKENISER_OP_TOKEN;
  else if (E_TOKENISER_CHAR_IS_WORD(tp,c)) tp->type = E_TOKENISER_WORD_TOKEN;
  else goto error_Invalid_Char;  
  loop = 1;

  while (loop)
  { tp->Buff[i++] = c;
    c = E_CHARSTREAM_GET(csp);
    loop = (((tp->type == E_TOKENISER_OP_TOKEN)   && 
              E_TOKENISER_CHAR_IS_OP(tp,c)       ) ||
            ((tp->type == E_TOKENISER_WORD_TOKEN) && 
              E_TOKENISER_CHAR_IS_WORD(tp,c))    );
  }
  E_CHARSTREAM_UNGET(csp,c);

  tp->Buff[i] = 0;  
  return(tp->type);

EndOfFile:
   tp->type = E_TOKENISER_EOF_TOKEN;
   return(-1);

error_EOF:
   e_error_syntax(
  "Line %d : Unexpected end of line while parsing a string starting line %d\n",
        tp->line_no,start);
   return(0);

error_Invalid_Char:
   e_error_syntax("Line %d: Invalid character %c\n",tp->line_no,(char)c);
   return(0);

error_Bad_Quot_Char:
   e_error_syntax("Line %d : Unterminated character constant\n",
           tp->line_no);
   return(0);

}

/******************* TOKENISER TOOLSS  ************************/

/**************************************************/
/* Set the char type to t for each character in s */
/* Range of char can be indicated as a-z.         */
/* -- is -                                        */
/* Return : 1 OK                                  */
/*          0 OK                                  */
/**************************************************/
int e_tokeniser_set_char_types(char *types,char *s,int t)
{ int i,j;
  unsigned int c,c2,nc,lastc;
  char *name=0;

  lastc = 0xfff;
  i = 0;
  while((c = s[i++]))
  { if((c == '-'))
    { nc = s[i++];
      if(nc == '-')         /*  "--" => - */
      { c2 = '-'; 
      }
      else if (nc == 0)  /*  "a-" => a to 255 */
      { c = lastc; c2 = 0xff; i--; lastc = 0xfff;
      }
      else               /*  "a-z" => a to z */
      { c = lastc; c2 = nc; lastc = 0xfff;
      }
    }
    else c2 = c;
    lastc = c;

    if((c > c2)||(c > 0xff) || (c2 > 0xff))
    { switch(t)
      { case E_TOKENISER_WORD_CHAR   : name = "Word Char"; break;
        case E_TOKENISER_OP_CHAR     : name = "Operator Char"; break;
        case E_TOKENISER_PUNCT_CHAR  : name = "Punctuation Char"; break;
        case E_TOKENISER_SPACE_CHAR  : name = "Space Char"; break;
      default : name ="UNKNOWN"; break;
      }
      e_error(E_ERROR_SYNTAX,E_ERROR_LEVEL_USER,
              "Invalid tokeniser %s list :\"%s\" at char %d",name,s,i);
      return(0);
    }

    for(j = c; j <= c2; j++)
    { types[j] = t;
    }
  }

  return(1);
}

/****************************************************/
/* Create a string to list al chars of a given type */
/* types : arays of char types                      */
/* buff : char [E_LINE_SIZE] : to store the string  */
/* t : char type to extract                         */
/* Range of char are converted to a-z.              */
/* -- is -                                          */
/****************************************************/
void e_tokeniser_char_types_to_buffer(char *types,char *buff,int t)
{ int i,j,k;
  unsigned int c,firstc;

  firstc = 0xfff;

  i = j = 0;
  while(i < 256)
  { if((c = types[i++]) == t)
    { if(firstc > 0xff)      /* first char of that type */
      { if((firstc = c) == '-')
        {  /* write - and restart after the - */
           buff[j++] = '-'; buff[j++] = '-'; firstc = 0xfff; 
        }
      } 
    
    }
    else if (firstc < 0xff)  /* end of sequence of chars of type t */
    { c--; /* last char of correct type : first c is never - */
      if(c == '-') /* push it back */
      { c--; 
      }

      if((c-firstc) > 3)  /* a-z*/
      { j += e_tokeniser_escape_char(firstc,buff + j);
        buff[j++] = '-';
        j += e_tokeniser_escape_char(c,buff + j);
        firstc = 0xfff;
      }
      else
      { for(k=firstc; k <= c; k++) 
        { if(k == '-') { buff[j++] = '-'; buff[j++] = '-';}
          else j += e_tokeniser_escape_char(k,buff + j);
	}
      }
    }
  }
}

/********************************************************/
/* Write the character c in s quoting the special chars */
/* Return : number of character written                 */
/* Example :                                            */
/* char Buff[100];                                      */
/*  i +=  e_tokeniser_escape_char(c,Buff+i);            */
/********************************************************/
int e_tokeniser_escape_char(int c,char *s)
{ int i;

  i = 0;
  switch(c) 
  { case '\a' : s[i++] = '\\'; s[i++] = 'a'; break;
    case '\b' : s[i++] = '\\'; s[i++] = 'b'; break;
    case '\t' : s[i++] = '\\'; s[i++] = 't'; break;
    case '\n' : s[i++] = '\\'; s[i++] = 'n'; break;
    case '\v' : s[i++] = '\\'; s[i++] = 'v'; break;
    case '\f' : s[i++] = '\\'; s[i++] = 'f'; break;
    case '\r' : s[i++] = '\\'; s[i++] = 'r'; break;
    default   : s[i++] = c; break;
  }

  return(i);
}
