#include <easy.h>
#include <ctype.h>

char *e_tools_unix_match_set(char *pat,char c);

/*****************************************************/
/* Test if the unix pattern pat matches the string s */
/* Pattern : * match anything.                       */
/*           . match one char                        */
/*           [ab-x] : match group of character       */
/* Return : 1 : match                                */
/*          0 : no match                             */
/*****************************************************/
int e_tools_unix_match(char *pat,char *s)
{ char c,pc;

  while((c = *s++))
  { pc = *pat++;

    switch(pc)
    { case 0   : return(0); break;
      case '.' : break; /* match anything */
      case '*' : if(e_tools_unix_match(pat,s-1)) return(1);   /* pat+1 s   */
                 if(e_tools_unix_match(pat,s)) return(1);   /* pat+1 s+1 */
                 if(e_tools_unix_match(pat-1,s)) return(1); /* pat   s+1 */
                 return(0);
      case '[' :  /* match set of character */
                  if(!(pat = e_tools_unix_match_set(pat,c))) return(0);
                  break;
      case '\\' : pc = *pat++; /* escape char */                  
      default: if(c != pc) return(0);
               break;
    }
  }
  if(*pat == 0) return(1); /* the 2 eos match */ 
  if ((*pat == '*') && (pat[1] == 0)) return(1);
  return(0); /* no match */

}

/************************************/
/* Match a block of character       */
/* [abA-Z]: ab or A to Z            */
/* [-z...] : 0 to z ...             */
/* [a-] : a to 255                  */
/* return : 0 no match              */
/* pat pointer just after ] : match */
/************************************/        
char *e_tools_unix_match_set(char *pat,char c)
{ char pc,last,loop;

  last = 0;
  loop = 1;
  while(loop && (pc = *pat++))
  { switch(pc)
    { case ']' : return(0); /* no match */
      case '-' : if ((*pat==0)||(*pat == ']'))
                 { if(last <= c) loop = 0;
                 }
                 else if ((last <= c) && (c <= *pat)) 
                 { loop = 0; 
                 }
                 break;
      case '\\' : pc = *pat++;
      default : if (pc == c) 
                { loop = 0;
                }
    }
    last = pc;
  }
  if(loop) return(0); /* no match */

  /* search for then end of block! */
  while ((pc = *pat++) && (pc != ']'));
  return(pc ? pat : pat-1);
}

/******************************************/
/* Trim the end of the string by removing */
/* all the characters listed in list.     */
/* str : string to trim.                  */
/* list : list of character to remove,    */
/*        if NULL remove all char < ' '   */
/* RETURN ; the size of the new string    */
/******************************************/
int e_tools_trim_eos(char *str,char *list)
{ int index,stop;
  
  if(!str) return(0);
  index = strlen(str)-1;

  stop = 0;

  while(!stop && (index >= 0))
  { if(list)
    { if(strchr(list,str[index] )) index--;
      else                         stop = 1;
    }
    else
    { if(str[index] <= ' ') index--;
      else                  stop=1;
    }
  }
  str[++index] = 0;
  return(index);
}

/************************************************************/
/* Remove string i from the NULL terminated list of string  */
/* l : NULL terminated array of string                      */
/* i : index of the item to remove (first is 0)             */
/* NOTE: if the removed string was dynamicaly allocated,    */
/* it will not be freed!                                    */
/* return : the new index position : i or i-1 if i was last */
/************************************************************/
int e_tools_del_strlist_item(char **l,int i)
{ int j;

  j = i;
  while(l[j])
  { l[j] = l[j+1]; 
    j++;
  }
  if(l[i] == 0) i--;
  if (i < 0) i = 0; /* for an empty list i is 0 */
  return(i);
}

/*********************************************************************/
/* Add string s at position i to the NULL terminated list of strings */
/* l    : NULL terminated array of string                            */
/* item : the string  to add to the list                             */
/* i    : index for the item.                                        */
/* size : the size of the array of string                            */
/* return : the new index position : i                               */
/*********************************************************************/
int e_tools_add_strlist_item(char **l,char *item,int i,int size)
{ int j;
  char *s,*d;

  /* ship every item up */
  j = i;
  s = l[j];
  while(s && (j < size-1))
  { d = l[j+1];
    l[j+1] = s;
    s = d; 
    j++;
  }
  l[j+1] = 0;

  l[i] = item;

  return(i);
}

/************************************************************/
/* Convert the string to an array of string                 */
/* s : the string to parse (is modified!)                   */
/* l : NULL terminated array of string                      */
/* sep : The list of char used as separators.               */
/* size the size of the array.                              */
/* If the array is too small some lines will be lost.       */
/* The string s must be a buffer as the separator char      */
/*  becomes NULL chars                                      */
/* Return : the number of lines                             */
/************************************************************/
int e_tools_str_to_str_list(char *s,char **l,char *sep,int size)
{ int i,j,c;
  char *first;

  if(!sep) sep = "\n\r";

  first = s;
  i = j = 0;
  while((c = s[i])&&(j < (size-1)))
  { if(strchr(sep,c))
    { s[i] = 0;
      l[j++] = first;
      first = s+i+1;
    }
    i++;
  } 
  if((first-s) < i) l[j++] = first;

  l[j] = 0;
  return(j);
}

/**********************************************/
/* check if the string s is an integer number */
/* return  : 1 s is an integer                */
/*           0 s is not an integer            */
/**********************************************/
int e_tools_is_an_int(char *s)
{ char c;

  if (!s) return(0);

  while((c = *s++))
  { if((c < '0') || ('9' < c)) return(0);
  }
  return(1);
}

/**********************************************/
/* check if the string s is a real number     */
/* return  : 1 s is a real                    */
/*           0 s is not a real                */
/**********************************************/
int e_tools_is_a_real(char *s)
{ double d;
  char *end;
 
  if (!s) return(0);

  d = strtod(s,&end);

  if((end == s) || (*end != 0)) return(0); 

  return(1);
}

/*********************************************************/
/* check if the string s has no space: no char <= SPACE  */
/* return  : 1 s is an integer                           */
/*           0 s is not an integer                       */
/*********************************************************/
int e_tools_has_no_spaces(char *s)
{ char c;

  if (!s) return(0);

  while((c=*s++))
  { if(c <= ' ') return(0);
  }
  return(1);
}

/****************************************************/
/* Copy s into buff, starting at position index     */
/* Ensure that the buffer size is not exceeded      */
/* The result is always a null terminated string    */
/* buff : destination buffer                        */
/* s    : string to copy                            */
/* index : index of the insertion position          */
/* size : buffer size                               */
/* return : index position of the end of string EOS */
/* Usage :                                          */
/*        i = 0;                                    */
/*        i = str_add(buff,s1,i,256);               */
/*        i = str_add(buff,s2,i,256);               */
/****************************************************/
int e_tools_str_add(char *buff,char *s,int index,int size)
{ int i;

  i = 0;
  size--;
  while((index < size) && s[i])
  { buff[index++] = s[i++];
  }
  buff[index] = 0;

  return(index);
}

/*************************************************/
/* Get word col from src and write it in dst     */
/* src : source buffer                           */
/* dst : destination buffer                      */
/* sep : list of separator chars " \t\n" if NULL */
/* col : index of word to extract (1st is 1)     */
/* size : max size of dst                        */
/* RETURN : size of word extracted               */
/*          0 : not that many word in the src    */
/*************************************************/
int e_tools_get_word(char *src, char *dst,char *sep,int col,int size)
{ int status,is_sep,i,j,k,pos,c,s;
#define TOOLS_SEP  0
#define TOOLS_WORD 1

  if(!sep) sep = " \t\n";

  status = TOOLS_SEP;
  pos = i = k = 0;
  while((c=src[i])!= 0)
  { j = 0;
    is_sep = 0; 
    while((s=sep[j++]) != 0 ) { if (c == s) is_sep = 1; }

    if(status == TOOLS_SEP)   /* new word */
    { if(!is_sep) 
      { status = TOOLS_WORD; 
        pos++;
      }
    }
    
    if(status == TOOLS_WORD) 
    { if(!is_sep)
      { if((col == pos) &&(k < (size-1))) dst[k++] = c;
      }
      else  status = TOOLS_SEP;
    }
    i++;
  }

  dst[k] = 0;

  return(k);
}


/******************************************************/       
/* Substitue the substring search by subst in src and */
/* put the result in dst. Do up to nb bubstitutions   */
/* src : source buffer                                */               
/* dst : output buffer                                */               
/* search : string to replace                         */               
/* subst : substitution string                        */               
/* size : maximum size for the dst buffer             */               
/* nb : no of substitutions (< 0 means no limit)      */    
/* Return : size of the new string                    */ 
/*          -1 : ERROR                                */   
/******************************************************/       
int e_tools_str_subs(char *src,char *dst,char *search,char *subst,int size, 
                     int nb)
{ char *p,*np;
  int i,si,di,lsi,sindex,search_len,subs_co;

  if(!src || !dst || !search || !subst) return(-1); 

  search_len = strlen(search);
  subs_co = 0;

  di = 0;
  sindex = 0;
  p = src;
  while((np = strstr(p,search)))
  { lsi = (np-src);
    si = sindex;
    /* copy what comes before search */
    while((si<lsi)&&(di < (size-1))) { dst[di++] = src[si++]; }

    /* copy subst */
    i = 0;
    if((nb < 0) || (subs_co < nb)) 
    { while((dst[di] = subst[i++])&&(di < (size-1))) di++;
      subs_co++;
    }
    else
    { while((dst[di] = search[i++])&&(di < (size-1))) di++;
    }

    /* skip search in src */
    p = np + search_len;
    sindex = p-src;
  }
  /* copy end of line */
  while((dst[di] = *p++)) {di++;}

  return(di);
} 

/************************************/
/* Convert the buffer to upper case */
/* Return: src                      */
/************************************/
char *e_tools_str_to_upper(char *src)
{ 
  if(!src) return(NULL);

  while(*src) { *src = toupper(*src); src++;}

  return(src);
}

/************************************/
/* Convert the buffer to lower case */
/* Return: src                      */
/************************************/
char *e_tools_str_to_lower(char *src)
{ 
  if(!src) return(NULL);

  while(*src) { *src = tolower(*src); src++;}

  return(src);
}

/*******************************************************/
/* Return the index of the string suffix:              */
/* return index of first char after the last sep and 0 */
/* if sep is not present in src                        */
/* src : string to scan                                */
/* sep : suffix separator                              */
/* Return: index position of suffix                    */
/*         0 if src is NULL                            */
/*******************************************************/
int e_tools_str_get_suffix(char *src,char sep)
{ int n;

  if(!src) return(0);

  n = strlen(src);

  while(n > 0 && (src[n-1] != sep)) n--;
  
  return(n);
}

/****************************************************/
/* Convert an hexadecimal char '0'->'f' or '0'->'F' */
/* to the integer varule (0 -> 15)                  */
/* Return the digit value (0 -> 15)                 */
/* -1 : error                                       */
/****************************************************/
int e_tools_hex_char_to_number(char c)
{ 
  if(('a' <= c) &&(c <= 'f')) return(c+10-'a');
  if(('A' <= c) &&(c <= 'F')) return(c+10-'A');
  if(('0' <= c) && (c <= '9')) return(c - '0');
  return(-1);
}

/****************************************/
/* Convert an octal char '0'->'7'       */
/* to the integer varule (0 -> 7)       */
/* Return the digit value (0 -> 7 )     */
/* -1 : error                           */
/****************************************/
int e_tools_octal_char_to_number(char c)
{ 
  if(('0' <= c) && (c <= '7')) return(c - '0');
  return(-1);
}

/**************************************************/
/* Convert a string with escape sequence          */
/* to a string with the corresponding char values */
/* "\n" -> \n ...                                 */
/* src : source string                            */
/* dst : destination string                       */
/* size : dst buffer size                         */
/* RETURN : 0 OK                                  */
/*          N : error: index position of          */
/*              the syntax error                  */
/**************************************************/
int e_tools_escape_str(char *src,char *dst,int size)
{ int c,i,j,state,d1,d2;

#define NORMAL 1
#define ESCAPE 2

  state = NORMAL;
  size --;

  i = j = 0;
  while((c= src[i++])&&(j < size))
  { if((state == NORMAL) && (c =='\\')) state = ESCAPE;
    else if (state == ESCAPE)
    { state = NORMAL;
      if(c=='x') 
      { if((d1 = e_tools_hex_char_to_number(src[i++])) < 0) return(i-1);
        if((d2 = e_tools_hex_char_to_number(src[i++])) < 0) return(i-1);
        c = d1*16+d2;
      }
      else if (('0' <= c)&&(c <= '3'))
      { if((d1 = e_tools_octal_char_to_number(src[i++])) < 0)
	{ if(c != '0') return(i-1); 
	    /* else \0 => 0 */
	  c = 0;
          i--;
        }
        else
        { if((d2 = e_tools_octal_char_to_number(src[i++])) < 0) return(i-1);
          c = (c-'0')*64+d1*8+d2;
        }
      } 
      else 
      { switch(c) 
        { case 'a' : c = '\a'; break;
          case 'b' : c = '\b'; break;
          case 'f' : c = '\f'; break;
          case 'n' : c = '\n'; break;
          case 'r' : c = '\r'; break;
          case 't' : c = '\t'; break;
          case 'v' : c = '\v'; break;
          case '0' : c = 0; break;
          case '\\' : c = '\\'; break;
        }
      }
      dst[j++] = c;
    }
    else dst[j++] = c;
  }
  dst[j] = 0;

  return(0);
}

/*********************************************/
/* Read one line from fpt into Buff          */
/* Remoce the trailing "\n"                  */
/* fpt : file opened for reading             */
/* Buff : destination buffer                 */
/* size : size of buffer                     */
/* return : size of line read (strlen(Buff)) */
/*          EOF : Buff empty end EOF reached */
/*********************************************/
int e_tools_read_line(FILE* fpt,char *Buff,int size)
{ int c,i;

  i = 0;
  while((c = fgetc(fpt)) && (c != '\n') && (c != EOF) && (i < size-1))
  { Buff[i++] = c;
  }
  Buff[i] = 0;

  if(!i && (c==EOF)) return(EOF);

  return(i);
}

/**
 *   void e_tools_print_aset(char *str,e_aset *asp, FILE *fp)
 *                  [.,0;.]
 **/
/*****************************************/
/* Print the aset with a prefix string : */
/* If str = "aset = ", the output is     */
/* aset = { ... }                        */
/* str : prefix string                   */
/* asp : aset to print                   */
/* fp : file pointer (0 -> stdout)       */
/* Link : asp : 0                        */
/*****************************************/
void e_tools_print_aset(char *str,e_aset *asp, FILE *fp)
{ e_string *e_sp;

  if(!fp) fp = stdout;

  if((e_sp=e_aset_to_string(asp,0,0)))
  { fprintf(fp,"%s%s\n",str,E_STRING_S(e_sp));
  E_SAFE_FREE(e_sp);
  }
}

/**
 *   void e_tools_print_obj(char *str,e_obj *obj, FILE *fp)
 *                  [.,0;.]
 **/
/*******************************************/
/* Print the object with a prefix string:  */
/* If str = "obj = ", the output is        */
/* obj = { ... }                           */
/* str : prefix string                     */
/* asp : aset to print                     */
/* fp : file pointer (0 -> stdout)         */
/* Link : asp : 0                          */
/*******************************************/
void e_tools_print_obj(char *str,e_obj *obj, FILE *fp)
{ e_string *e_sp;

  if(!fp) fp = stdout;

  if((e_sp=e_aset_to_string(obj,0,0)))
  { fprintf(fp,"%s%s\n",str,E_STRING_S(e_sp));
    E_SAFE_FREE(e_sp);
  }

}
