#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <easy.h>
#include <sysfrpm.h>
#include "sysfrpm.BUILD"

#define LIST_OVERWRITE     0 
#define LIST_NO_OVERWRITE  1
#define LIST_MULTI_KEY     2
#define LIST_MATCH_VAL     4
#define LIST_REMOVE       32

#define CMD_DIFF      1
#define CMD_CHECK     2
#define CMD_CHECK_SRC 4
#define CMD_DIFF_LIST 8

#ifdef LINUX
extern int gethostname (char *__name, size_t __len);
#endif

/**************************************************************/ 
/* sysf : Unix/Linux system file management system            */
/**************************************************************/ 

char VERSION[] = "sysfrpm: v1.01."BUILD_NO"; "BUILD_DATE";  Bernard Piette";

/* Define some Text Constants */
char *Help[] = { 
"Syntax : sysfrpm SRC_LIST [INST_LIST] [COMMAND] [FLAGS]",
" SRC_LIST: [-name=n NAME] [-dir=d NAME] [-MasterFile=mf NAME] [-installed=i]",
"  name=n : read rpm list from the file.",
"  dir=d : scan directory for .rpm files.",
"  MasterFile=mf : file is list of files and directory to scan.",
"  installed=i : scan the list of installed rpm (rpm -qa)",
" INST_LIST: [-InstName=in NAME] [-InstDir=id NAME] [-InstMasterFile=imf NAME]",
"  DEFAULT : installed rpms",
"  InstName=in : read rpm list from the file.",
"  InstDir=id : scan directory for .rpm files.",
"  InstMasterFile=imf : file is list of files and directory to scan.",
" COMMAND: [-diff=D] [-check=C] [-CheckSrc=CS] [-DiffList=DL]",
"  DEFAULT : print the ordered list of rpm in SRC_LIST",
"  diff=D : compare rpm in SRC_LIST to rpm in INST_LIST",
"           List Missing, Extra or rpm with wrong versions", 
"  check=C : compare rpm in SRC_LIST to rpm in INST_LIST with no output.",
"            exit(0) if no difference, else exit(1)",
"  CheckSrc=CS : Check that the SRC_LIST are properly instaled.", 
"  DiffList=DL : print as an ordered list the difference between SRC_LIST and",
"           INST_LIST",
"  RemList=RL : Print list of extra rpm on one line (for rpm -e)",
" FLAGS: [NoMissing=nm] [NoExtra=ne] [NoWrongVer=nw]",
"  NoMissing=nm : do not display the missing packages.",
"  NoExtra=ne : do not display the extra packages.",
"  NoWrongVer=nw : do not display the packages with a version mismatch.",
0};

struct {
  char MasterFile[E_LONG_BUFF_SIZE];
  char name[E_LONG_BUFF_SIZE];
  char dir[E_LONG_BUFF_SIZE];
  char InstMasterFile[E_LONG_BUFF_SIZE];
  char InstName[E_LONG_BUFF_SIZE];
  char InstDir[E_LONG_BUFF_SIZE];
  int installed;
  int diff;
  int check;
  int CheckSrc;
  int DiffList;
  int RemList;
  int NoMissing;
  int NoExtra;
  int NoWrongVer;
} par;

e_parameter Pars[] = {
{ "MasterFile",E_CTYPE_STRING,par.MasterFile,"",0,"mf",0},
{ "name",E_CTYPE_STRING,par.name,"",0,"n",0},
{ "dir",E_CTYPE_STRING,par.dir,"",0,"d",0},
{ "InstMasterFile",E_CTYPE_STRING,par.InstMasterFile,"",0,"imf",0},
{ "InstName",E_CTYPE_STRING,par.InstName,"",0,"in",0},
{ "InstDir",E_CTYPE_STRING,par.InstDir,"",0,"id",0},
{ "installed",E_CTYPE_FLAG,&par.installed,0,1,"i",0},
{ "diff",E_CTYPE_FLAG,&par.diff,0,1,"D",0},
{ "check",E_CTYPE_FLAG,&par.check,0,1,"C",0},
{ "CheckSrc",E_CTYPE_FLAG,&par.CheckSrc,0,1,"CS",0},
{ "DiffList",E_CTYPE_FLAG,&par.DiffList,0,1,"DL",0},
{ "RemList",E_CTYPE_FLAG,&par.RemList,0,1,"RL",0},
{ "NoMissing",E_CTYPE_FLAG,&par.NoMissing,0,1,"nm",0},
{ "NoExtra",E_CTYPE_FLAG,&par.NoExtra,0,1,"ne",0},
{ "NoWrongVer",E_CTYPE_FLAG,&par.NoWrongVer,0,1,"nw",0},
{ 0,0,0,0,0,0},
};

char *rpm_qa[] ={"rpm","-qa",
  "--queryformat","%{NAME}-%{VERSION}-%{RELEASE}.%{ARCH}\n",0};
e_aset *RPM_list,*RPM_inst;
e_aset *RPM_missing,*RPM_wrongv,*RPM_extra;

char HostName[E_LINE_SIZE],HOSTNAME[E_LINE_SIZE];

int parse_rpm_filename(char *name,rpm_info *rpmi);
int parse_rpm_name(char *name,rpm_info *rpmi);
int read_rpm_master_file(char *file_name,e_aset **rpmp);
int read_rpm_filelist(char *file_name,e_aset **rpmp);
int read_rpm_dirlist(char *file_name,e_aset **rpmp);
int read_installed_rpm();
e_aset *make_rpm_aset(rpm_info *rpmi);
e_aset *rpml_add(e_aset *rpmlp,rpm_info *rpmi,int flag);
e_aset *rpml_del(e_aset *rpmlp,rpm_info *rpmi,int flag);
int rpm_compare_lists();
int print_aset_rpm_list(e_aset *rpmlp,FILE *fpt);
int print_aset_rpm_one_line(e_aset *rpmlp,FILE *fpt);
int check_src();


/**************************************************************************/
/*                        MAIN                                            */   
/**************************************************************************/
int main(argc, argv)
int argc;
char *argv[];
{ int Argc,err,command;
  char **Argv;

  E_NULL_ERROR(e_init());

           /* Read the argument list*/
           /*************************/
  if((Argc = e_parameter_read_arg_line(argc,argv,Pars,Help,&Argv,
                                       E_PARAMETER_V_IS_VERSION))<0)
  { goto error;
  }

  /* get host name */
  if(gethostname(HostName,E_LINE_SIZE))
  { fprintf(stderr,"Unable to get host name!\n");
    goto error;
  }
  strcpy(HOSTNAME,HostName);
  e_tools_str_to_upper(HOSTNAME);

  /* set command flag */
  command = 0;
  if(par.diff || par.RemList) command |= CMD_DIFF;
  if(par.check) command |= CMD_CHECK;
  if(par.CheckSrc) command |= CMD_CHECK_SRC;
  if(par.DiffList) command |= CMD_DIFF_LIST;
  
  /* Initialise the RPM aset lists */
  if(!(RPM_list = e_aset_new(0))) goto error;
  if(!(RPM_inst = e_aset_new(0))) goto error;

  /* LOAD LIST */
  if(par.name[0] && !read_rpm_filelist(par.name,&RPM_list)) 
  { goto error;
  }

  if(par.dir[0] && !read_rpm_dirlist(par.dir,&RPM_list)) 
  { goto error;
  }

  if(par.MasterFile[0] && !read_rpm_master_file(par.MasterFile,&RPM_list)) 
  { goto error;
  }

  if(par.installed && !read_installed_rpm(&RPM_list)) goto error;

  E_NULL_ERROR((RPM_list = 
	       e_aset_sort_by_key(RPM_list,E_MODE_NOCLONE|E_MODE_INCREASING)));

  /* LOAD SRC (INSTALLED) LIST */
  if(command)
  { if(par.InstName[0]) 
    { E_NULL_ERROR(read_rpm_filelist(par.InstName,&RPM_inst));
    }
    else if(par.InstDir[0])
    { E_NULL_ERROR(read_rpm_dirlist(par.InstDir,&RPM_inst));
    }
    else if(par.InstMasterFile[0]) 
    { E_NULL_ERROR(read_rpm_master_file(par.InstMasterFile,&RPM_inst));
    }   
    else
    { E_NULL_ERROR(read_installed_rpm(&RPM_inst));
    }

    E_NULL_ERROR((RPM_inst = 
	       e_aset_sort_by_key(RPM_inst,E_MODE_NOCLONE|E_MODE_INCREASING)));
  }


  /***************** DIFF list and installed ******************/
  if(par.diff)
  { rpm_compare_lists();
    err = 0;

    if(!par.NoMissing && E_ASET_CARDINAL(RPM_missing))
    { printf("# MISSING RPM :\n");
      print_aset_rpm_list(RPM_missing,stdout);
      printf("\n");
      err += E_ASET_CARDINAL(RPM_missing);
    }

    if(!par.NoExtra && E_ASET_CARDINAL(RPM_extra))
    { printf("# EXTRA RPM :\n");
      print_aset_rpm_list(RPM_extra,stdout);
      printf("\n");
      err += E_ASET_CARDINAL(RPM_extra);
    }

    if(!par.NoWrongVer && E_ASET_CARDINAL(RPM_wrongv))
    { printf("# WRONG VERSION RPM :\n");
      print_aset_rpm_list(RPM_wrongv,stdout);
      err += E_ASET_CARDINAL(RPM_wrongv);
    }

    exit(err ? 1 : 0);

  }
  /***************** CHECK list and installed ******************/
  else if(par.CheckSrc)
  { if((err = check_src()) < 0) goto error;

    if(!par.NoMissing && E_ASET_CARDINAL(RPM_missing))
    { printf("# MISSING RPM :\n");
      print_aset_rpm_list(RPM_missing,stdout);
      printf("\n");
      err += E_ASET_CARDINAL(RPM_missing);
    }

    if(!par.NoWrongVer&& E_ASET_CARDINAL(RPM_wrongv))
    { printf("# WRONG VERSION RPM :\n");
      print_aset_rpm_list(RPM_wrongv,stdout);
      err += E_ASET_CARDINAL(RPM_wrongv);
    }

    exit(err ? 1 : 0);

  }
  /***************** CHECK list and installed ******************/
  else if(par.check)
  { rpm_compare_lists();

    if((E_ASET_CARDINAL(RPM_missing) > 0) || 
       (E_ASET_CARDINAL(RPM_extra) > 0) || 
       (E_ASET_CARDINAL(RPM_wrongv) > 0)) 
    { exit(1);
    }
    else exit(0);

  }
  /***************** MAKE DIFF LIST ******************/
  else if(par.DiffList)
  { E_NULL_ERROR(RPM_list = e_aset_diff_by_key(RPM_list,RPM_inst,-1));

    print_aset_rpm_list(RPM_list,stdout);

  }
  /***************** MAKE REM LIST ******************/
  else if(par.RemList)
  { rpm_compare_lists();

    if(RPM_extra && E_ASET_CARDINAL(RPM_extra))
    { print_aset_rpm_one_line(RPM_extra,stdout);
    }
    printf("\n");

  }
  /***************** MAKE LIST ******************/
  else
  {  print_aset_rpm_list(RPM_list,stdout);
  }

  exit(0);

error:
  e_error_flush(E_ERROR_ANY);
  exit(1); 
}

/****************************** PARSE **************************************/

/*************************************************************************/
/* Parse an RPM file name and extract the name, version and architecture */
/* name : string to parse                                                */
/* rpmi : rpm_info for the result                                        */
/* RETURN : 1 : OK                                                       */
/*          0 : ERROR                                                    */
/*************************************************************************/
int parse_rpm_filename(char *name,rpm_info *rpmi)
{ int i,j,size;

  size = strlen(name);

  /* .rpm filename ? if so remove the suffix */
  if(strcmp(&name[size-4],".rpm")) return(0);
  
  size -= 4;

  /* extract the architecture */
  i = size-1;
  while ((i >= 0)&& (name[i] != '.')) i--;
  if(i < 0) return(0);

  for(j=0; j < (size-i-1); j++)
  { rpmi->arch[j] = name[i+j+1];
  }
  rpmi->arch[j] = 0;  
  size = i;
 
  /* extract the version no */
  while ((i >= 0)&& (name[i] != '-')) i--;
  if(i < 0) return(0);
  i--;
  while ((i >= 0)&& (name[i] != '-')) i--;
  if(i < 0) return(0);

  for(j=0; j < (size-i-1); j++)
  { rpmi->version[j] = name[i+j+1];
  }
  rpmi->version[j] = 0;  
  size = i;
   
  /* extract the RPM name */
  for(j=0; j < i; j++)
  { rpmi->name[j] = name[j];
  }
  rpmi->name[j] = 0;

  return(1);
}

/*************************************************************************/
/* Parse an RPM name (as output by rpm -qa) and extract the name,        */
/* version and arch                                                      */
/* name : string to parse                                                */
/* rpmi : rpm_info for the result                                        */
/* RETURN : 1 : OK                                                       */
/*          0 : ERROR                                                    */
/*************************************************************************/
int parse_rpm_name(char *name,rpm_info *rpmi)
{ int i,j,size;

  size = strlen(name);

  /* extract the architecture */
  i = size-1;
  while ((i >= 0)&& (name[i] != '.')) i--;
  if(i < 0) return(0);

  for(j=0; j < (size-i-1); j++)
  { rpmi->arch[j] = name[i+j+1];
  }
  rpmi->arch[j] = 0;  
  size = i;

  /* extract the version no */
  while ((i >= 0)&& (name[i] != '-')) i--;
  if(i < 0) return(0);
  i--;
  while ((i >= 0)&& (name[i] != '-')) i--;
  if(i < 0) return(0);

  for(j=0; j < (size-i-1); j++)
  { rpmi->version[j] = name[i+j+1];
  }
  rpmi->version[j] = 0;  
  size = i;
   
  /* extract the RPM name */
  for(j=0; j < i; j++)
  { rpmi->name[j] = name[j];
  }
  rpmi->name[j] = 0;

  return(1);
}

/************************************************************/
/* Return a pointer to the first non space character in str */
/************************************************************/
char *name_start(char *str)
{ 
  while(str && (*str == ' ')&&(*str == '\t')&&(*str == '\n')) str++;
  return(str); 
}

/****************************** READ RPM LISTS ******************************/

/******************************************************************/
/* Extract a list of rpm file from the master file :              */
/* MASTER FILE SYNTAX : one path per line. Directories            */
/*   are scanned for ".rpm" files and files must contains         */
/*   a list of packages                                           */
/* Parse each file and directory and add the names into RPM_list. */
/* RETURN : 1 : OK                                                */
/*          0 : ERROR                                             */
/******************************************************************/
int read_rpm_master_file(char *file_name,e_aset **rpmp)
{ FILE *fpt;
  char Buff[E_LONG_BUFF_SIZE],Buff2[E_LONG_BUFF_SIZE],*s,*path;
  struct stat stp;

  E_NULL_ERROR(fpt=e_open_file(file_name,"r"));

  while(e_tools_read_line(fpt,Buff,E_LONG_BUFF_SIZE) != EOF)
  { if(Buff[0] != '#') /* ignore comment lines */
    { /* $h  => hostname */
      /* $H  => HOSTNAME */
      e_tools_str_subs(Buff,Buff2,"$h",HostName,E_LONG_BUFF_SIZE,-1);
      e_tools_str_subs(Buff2,Buff,"$H",HOSTNAME,E_LONG_BUFF_SIZE,-1);
      e_tools_trim_eos(Buff," \t");
      s = name_start(Buff);

      path = s;
      if(*path == '-') path++;
      if(*path == '+') path++;
      if(*path == '@') path++;

      if(path[0] && !stat(path,&stp))
      { if(S_ISDIR(stp.st_mode))
        { E_NULL_ERROR(read_rpm_dirlist(s,rpmp));
        }
        else if(S_ISREG(stp.st_mode))
        { E_NULL_ERROR(read_rpm_filelist(s,rpmp));
        }
      } 
    }
  }
  return(1);  

error:
  return(0);
}

/***********************************************************/ 
/* Read an RPM file list : each line is the name of an RPM */
/* Lines starting with a # are ignored                     */
/* Print the name and version for each                     */
/* EVENTUALY WILL CREATE A LIST                            */
/* file_name : the file to read                            */
/* rpmp : pointer to aset to fill in.                      */
/* mode  : flag mode for rpml_add().                       */
/*         LIST_REMOVE : remove rpm from the list          */
/* RETURN : 1 : OK                                         */
/*          0 : ERROR                                      */
/***********************************************************/ 
int read_rpm_filelist(char *file_name,e_aset **rpmp)
{ rpm_info RPM_data;
  FILE *fpt;
  char Buff[256],*s;
  int mode,lmode;

  if(!file_name || !file_name[0]) return(1);

  if(file_name[0] == '-') /* remove rpm from the list*/
  { file_name++;
    mode = LIST_REMOVE;
    if(file_name[0] == '@') { file_name++; mode |= LIST_MATCH_VAL; }
  }
  else
  { mode = LIST_OVERWRITE;
    if(file_name[0] == '@') { file_name++; mode |= LIST_MULTI_KEY; }
  }

  E_NULL_ERROR(fpt=e_open_file(file_name,"r"));

  while(e_tools_read_line(fpt,Buff,256) != EOF)
  { e_tools_trim_eos(Buff," \t");
    s = name_start(Buff);
    lmode = mode;

    if(s[0] && (s[0] != '#'))
    { if(s[0] == '-') /* remove rpm from the list */
      { s++;
        lmode |= LIST_REMOVE;
      }
      else if (s[0] == '+') /* add rpm to the list */
      { s++;
        lmode &= ~LIST_REMOVE;
      }

      if (s[0] == '@') /* Multi key or match val */
      { s++;
        if(lmode & LIST_REMOVE) lmode |= LIST_MATCH_VAL; 
        else                    lmode |= LIST_MULTI_KEY;
      }

      if(parse_rpm_name(s,&RPM_data))
      { if(lmode & LIST_REMOVE)
        { if(!(*rpmp = rpml_del(*rpmp,&RPM_data,lmode))) goto error;
        }
        else
        { if(!(*rpmp = rpml_add(*rpmp,&RPM_data,lmode))) goto error;
        }
      }
    }
  }
  return(1);  

error:
  return(0);
} 

/***********************************************************/ 
/* Extract the list of RPM from a directory                */
/* Print the name and version for each                     */
/* EVENTUALY WILL CREATE A LIST                            */
/* file_name : the file to read                            */
/* rpmp : pointer to aset to fill in.                      */
/* mode  : flag mode for rpml_add().                       */
/* RETURN : 1 : OK                                         */
/*          0 : ERROR                                      */
/***********************************************************/ 
int read_rpm_dirlist(char *file_name,e_aset **rpmp)
{ rpm_info RPM_data;
  DIR *dpt;
  struct dirent *dent;
  int mode;

  if(!file_name || !file_name[0]) return(1);

  if(file_name[0] == '-') /* remove rpm from the list*/
  { file_name++;
    mode = LIST_REMOVE;
    if(file_name[0] == '@') { file_name++; mode |= LIST_MATCH_VAL; }
  }
  else
  { mode = LIST_OVERWRITE;
    if(file_name[0] == '@') { file_name++; mode |= LIST_MULTI_KEY; }
  }

  if(!S_ISDIR(e_stat_file(file_name)) || !(dpt = opendir(file_name)))
  { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
              "Can't open directory %s",file_name);
    goto error;
  }

  while((dent = readdir(dpt)) != NULL)
  { 
    if(parse_rpm_filename(dent->d_name,&RPM_data))
    {
      if(mode & LIST_REMOVE)
      { if(!(*rpmp = rpml_del(*rpmp,&RPM_data,mode))) goto error;
      }
      else
      { if(!(*rpmp = rpml_add(*rpmp,&RPM_data,mode))) goto error;
      }

    }
    /* else : not an RPM file and ignore it */
  }
  closedir(dpt);
  return(1);

error:
  return(0);
} 

/***********************************************************/ 
/* Extract the list of installed RPMs (rpm -qa)            */
/* rpmp : pointer to an aset pointer : aset to fill in     */
/* RETURN : 1 : OK                                         */
/*          0 : ERROR                                      */
/***********************************************************/ 
int read_installed_rpm(e_aset **rpmp)
{ rpm_info RPM_data;
  FILE *fpt;
  int c,i,sig,stat,pid,status;
  char BUFF[E_LONG_BUFF_SIZE],BUFF2[E_LONG_BUFF_SIZE];

  i = 0;
  if(!(pid = e_exec_pipe_fread(rpm_qa,&fpt))) goto error;  
  
  do
  { c = getc(fpt);

    if((((BUFF[i] = c)=='\n') || (c == EOF) || (i >= (E_LONG_BUFF_SIZE-1)))
       && (i>0))
    { BUFF[i] = 0;

      e_tools_get_word(BUFF,BUFF2,0,1,E_LONG_BUFF_SIZE);

      parse_rpm_name(BUFF2,&RPM_data);

      if(!(*rpmp = rpml_add(*rpmp,&RPM_data,LIST_MULTI_KEY))) 
           goto error;

      i = -1;
    }  
    i++;
  }
  while (c != EOF);

  fclose(fpt); 
  status = e_exec_wait_for(pid,&sig,&stat); 

  return(1);

error:
  fprintf(stdout,"error");
  return(0);
} 

/****************************************************************/
/* Compare 2 rpm                                                */
/* rpmi1, rpmi2 : rpm to compare                                */
/* RETURN : RPM_EQ : they are identical                         */
/*          RPM_DIFF_VER : same rpm name but different versions */
/*          RPM_DIFF_NAME : different name                      */
/****************************************************************/
int rpm_compare(rpm_info *rpmi1,rpm_info *rpmi2)
{
  if(strcmp(rpmi1->name,rpmi2->name)) return(RPM_DIFF_NAME);
  if(strcmp(rpmi1->version,rpmi2->version)) return(RPM_DIFF_VER);
  return(RPM_EQ);

}

/****************************** ASET **************************************/
/* Create an rpm asset :                                                  */
/* { name= "RPM_NAME"; version = "Version"; arch = "ARCH"}                */
/* Return : new aset.                                                     */
/*          0 : error                                                     */
/* LINK : RETURN : +1                                                     */
/**************************************************************************/
e_aset *make_rpm_aset(rpm_info *rpmi)
{ e_string *str_p; 
  e_aset *rpm_p=0;

  /* Create the new aset */
  if(!(rpm_p = e_aset_new(0))) goto error;

  /* Add the RPM name */
  if(!(str_p = e_string_new(0,rpmi->name,0))) goto error;
  if(!(rpm_p = e_aset_append_item(rpm_p,e_name_new("name"),str_p))) 
      goto error;

  /* Add the RPM version */
  if(!(str_p = e_string_new(0,rpmi->version,0))) goto error;
  if(!(rpm_p = e_aset_append_item(rpm_p,e_name_new("version"),str_p))) 
      goto error;

  /* Add the RPM arch */
  if(!(str_p = e_string_new(0,rpmi->arch,0))) goto error;
  if(!(rpm_p = e_aset_append_item(rpm_p,e_name_new("arch"),str_p))) 
      goto error;

  return(rpm_p);

error:
  if(rpm_p) E_FREE(rpm_p);
  return(0);
}


/*****************************************************************/
/* Add one rpm to the RPM_list                                   */
/* Each RPM is added to the rpmlp aset as a three element aset : */
/* (name; version; arch). The aset name (key) is the rpm name    */
/* rpmlp : rpm list aset                                         */
/* rpmi : rpm to add to the list                                 */
/* flag : LIST_OVERWRITE, LIST_NO_OVERWRITE                      */
/*        LIST_MULTI_KEY                                         */
/* RETURN : new RPM aset                                         */
/*          0 : ERROR                                            */
/* LINK : rpmlp : -1                                             */
/*        RETURN : +1                                            */
/*****************************************************************/
e_aset *rpml_add(e_aset *rpmlp,rpm_info *rpmi,int flag)
{ e_aset *rpm_p=0;
  int mode;

  if(flag & LIST_MULTI_KEY) mode = 0;
  else                      mode = E_ASET_ADD_MODE_UNIQUE_KEY;
  if(flag & LIST_NO_OVERWRITE) mode |= E_ASET_ADD_MODE_NO_REPLACE;

  E_NULL_ERROR(rpm_p = make_rpm_aset(rpmi));

  if(!(rpmlp = e_aset_add_item(rpmlp,e_name_new(rpmi->name),rpm_p,mode))) 
  { rpm_p = 0;  
    goto error;
  }

  return(rpmlp);

error:
  if(rpm_p) E_FREE(rpm_p);
  if(rpmlp) E_FREE(rpmlp);
  return(0);
}

/*****************************************************************/
/* Remove one rpm to the RPM_list                                */
/* rpmlp : rpm list aset                                         */
/* rpmi : rpm to add to the list                                 */
/* flag : LIST_MATCH_VAL or 0                                    */
/* RETURN : new RPM aset                                         */
/*          0 : ERROR                                            */
/* LINK : rpmlp : -1                                             */
/*        RETURN : +1                                            */
/*****************************************************************/
e_aset *rpml_del(e_aset *rpmlp,rpm_info *rpmi,int flag)
{ e_aset *rpm_p=0;
  e_name *key=0;
  int pos;

  E_NULL_ERROR(key = e_name_new(rpmi->name));

  if(flag & LIST_MATCH_VAL)
  { E_NULL_ERROR(rpm_p = make_rpm_aset(rpmi));
    if((pos = e_aset_get_item_index(rpmlp, key,rpm_p,0)) < 0) return(rpmlp);
  }
  else /* match key only */
  { if((pos = e_aset_get_key_index(rpmlp, key,0)) < 0) return(rpmlp);
  }

  key=0;

  E_NULL_ERROR(e_aset_remove_item_by_index(rpmlp,pos));

  return(rpmlp);

error:
  if(key) E_FREE(key);
  if(rpmlp) E_FREE(rpmlp);
  return(0);
}

/*****************************************************************/
/* Compare the RPM_list and the RPM_inst list of RPM             */
/* Build 3 new lists:                                            */
/* RPM_missing : rpm in the list but not installed               */
/* RPM_wrongv : rpm in the list installed but with grong version */
/* RPM_extra : rpm installed but not in the list                 */
/* RETURN : 1 : OK                                               */
/*          0 : error                                            */
/*****************************************************************/
int rpm_compare_lists()
{ e_aset_item *aip;
  e_aset *rpm_list=0;
  e_string *sp;
  void *rpm1,*rpm2,*wrong_rpm;
  e_name *key;
  int pos,next_pos,status,rpm_pos=0;

#define RPM_MISSING 1
#define RPM_GOOD    2
#define RPM_WRONGV  3

  E_NULL_ERROR(RPM_extra  = e_aset_clone(RPM_inst,E_CLONE_RECURSIVE));
  E_NULL_ERROR(rpm_list  = e_aset_clone(RPM_list,E_CLONE_RECURSIVE));
  E_NULL_ERROR(RPM_missing = e_aset_new(0));
  E_NULL_ERROR(RPM_wrongv = e_aset_new(0));
 
  aip = E_ASET_FIRST(rpm_list);

            /* Remove all the matching entries */
            /* Including Multiple ones         */
  while(aip)
  { status = RPM_MISSING;
    if((key = E_ASET_ITEM_NAME(aip)))
    { pos = 0; 
      next_pos = 0;
      wrong_rpm = 0;

      while(pos >= 0)
      { pos = e_aset_get_key_index(RPM_extra,E_USE(key),next_pos);
        next_pos = pos+1;
 
        if(pos >= 0)
        { rpm2 = e_aset_get_value_by_index(RPM_extra,pos);
          rpm1 = E_ASET_ITEM_VAL(aip);

/*
	sp = E_TO_STRING(rpm1,0,0);
        fprintf(stderr," rpm1=%s\n",E_STRING_S(sp));
        E_FREE(sp);
	sp = E_TO_STRING(rpm2,0,0);
        fprintf(stderr," rpm2=%s\n",E_STRING_S(sp));
        E_FREE(sp);
*/

          if(e_aset_equals(rpm1,rpm2))
          { status = RPM_GOOD;
            e_aset_remove_item_by_index(RPM_extra,pos);
            next_pos = pos;
          }
          E_FREE(rpm2); 
        }
      }
    }

    if(status == RPM_GOOD)
    { E_ASET_CUR(rpm_list) = aip;
      e_aset_remove_cur_item(rpm_list);
      aip = E_ASET_CUR(rpm_list);
    }
    else
    { aip = aip->next;
    }
  }

            /* Remove the entries with the wrong version */
            /* And make the wrong_rpm aset               */
  aip = E_ASET_FIRST(rpm_list);
  while(aip)
  { if((key = E_ASET_ITEM_NAME(aip)))
    { status = RPM_MISSING;
      pos = 0; 
      next_pos = 0;
      wrong_rpm = 0;

      while(pos >= 0)
      { pos = e_aset_get_key_index(RPM_extra,E_USE(key),next_pos);
        next_pos = pos+1;
 
        if(pos >= 0)
        { rpm2 = e_aset_get_value_by_index(RPM_extra,pos);
          rpm1 = E_ASET_ITEM_VAL(aip);

          if(!e_aset_equals(rpm1,rpm2))
	  { status = RPM_WRONGV;
            if(wrong_rpm) E_FREE(wrong_rpm);
            wrong_rpm = rpm2;
            rpm_pos = pos;
          }
          else 
          { sp = E_TO_STRING(rpm1,0,0);
            e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
              "sysfrpm inconsistency: double match for %s",E_STRING_S(sp));
            E_FREE(sp); 
            E_FREE(rpm2); 
            goto error;
          }
        }
      }

      switch(status)
      { case RPM_MISSING :
               /* missing rpm : add it to RPM_missing */
               E_NULL_ERROR(RPM_missing = 
                              e_aset_append_item(RPM_missing,E_USE(key),
                                          E_USE(E_ASET_ITEM_VAL(aip))));
               break;

        case RPM_WRONGV :
               E_NULL_ERROR(RPM_wrongv = 
                      e_aset_append_item(RPM_wrongv,E_USE(key),wrong_rpm));
               e_aset_remove_item_by_index(RPM_extra,rpm_pos);
               break;
      }
    }
    aip = aip->next;
  }

  E_SAFE_FREE(rpm_list); 
  return(1);

error:
  E_SAFE_FREE(rpm_list); 
  E_SAFE_FREE(RPM_extra); RPM_extra = 0;
  E_SAFE_FREE(RPM_missing); RPM_missing = 0;
  E_SAFE_FREE(RPM_wrongv); RPM_wrongv = 0;
  return(0);
}

/*****************************************/
/* Print the aset RPM list into the file */
/* rpmlp : aset of RPMs.                 */
/* fpt : file open for writting          */
/* return : 1 : OK                       */
/*          0 : ERROR                    */
/*****************************************/
int print_aset_rpm_list(e_aset *rpmlp,FILE *fpt)
{ e_aset_item *aip,*aip2;
  e_aset *rpmp;
  char *name,*version,*arch;

  aip = E_ASET_FIRST(rpmlp);

  while(aip)
  { /* get name and value */
    rpmp = E_ASET_ITEM_VAL(aip);
    aip2 = E_ASET_FIRST(rpmp);
    name = E_STRING_S(E_ASET_ITEM_VAL(aip2));
    aip2 = aip2->next;
    version = E_STRING_S(E_ASET_ITEM_VAL(aip2));
    aip2 = aip2->next;
    arch = E_STRING_S(E_ASET_ITEM_VAL(aip2));
      
    fprintf(fpt,"%s-%s.%s\n",name,version,arch);

    aip = aip->next;
  }

  return(1);
}

/*****************************************/
/* Print the aset RPM list on a single   */
/* as for use with rpm -e                */
/* rpmlp : aset of RPMs.                 */
/* fpt : file open for writting          */
/* return : 1 : OK                       */
/*          0 : ERROR                    */
/*****************************************/
int print_aset_rpm_one_line(e_aset *rpmlp,FILE *fpt)
{ e_aset_item *aip,*aip2;
  e_aset *rpmp;
  char *name,*version,*arch;

  aip = E_ASET_FIRST(rpmlp);

  while(aip)
  { /* get name and value */
    rpmp = E_ASET_ITEM_VAL(aip);
    aip2 = E_ASET_FIRST(rpmp);
    name = E_STRING_S(E_ASET_ITEM_VAL(aip2));
      
    fprintf(fpt,"%s ",name);

    aip = aip->next;
  }

  return(1);
}

/************************************************************************/
/* Check that the rpm listed in the SRC_list are properly installed:    */
/* Check that each element of RPM_list is in RPM_inst.                  */
/* Create two asets :                                                   */
/*   RPM_missing : the rpms in SRC_list that are not installed          */
/*   RPM_wrongv : the rpms in SRC_list installed with the wrong version */
/* RETURN : the number of missing+wrongv rpms.                          */
/*          -1 : ERROR                                                  */
/************************************************************************/
int check_src()
{ e_aset_item *aip;
  e_name *key;
  void *val;
  int n=0;

  E_NULL_ERROR(RPM_missing = e_aset_new(0));
  E_NULL_ERROR(RPM_wrongv = e_aset_new(0));

  aip = E_ASET_FIRST(RPM_list);
  
  while(aip)
  { key = E_ASET_ITEM_NAME(aip);
    val = E_ASET_ITEM_VAL(aip);

    switch(e_aset_is_in(RPM_inst,key,val))
    { case 0 : E_NULL_ERROR(RPM_missing = 
                      e_aset_append_item(RPM_missing,E_USE(key),E_USE(val)));
               n++;
               break;
      case -1 : E_NULL_ERROR(RPM_wrongv = 
                      e_aset_append_item(RPM_wrongv,E_USE(key),E_USE(val)));
               n++;
               break;
      case 1 : /* nothing to do */
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }

  return(n);

error:
  E_FREE(RPM_missing); RPM_missing = 0;
  E_FREE(RPM_wrongv); RPM_wrongv = 0;
  return(-1);
}
